import uuid
import ipaddress
import re

from marshmallow import Schema, fields, ValidationError, validates, validates_schema
from marshmallow.validate import Length, Range, OneOf

from vantage6.backend.common.resource.input_schema import (
    MAX_LEN_NAME,
    MAX_LEN_STR_LONG,
    validate_name,
)
from vantage6.common.task_status import TaskStatus
from vantage6.server.model.common.utils import validate_password

_MAX_LEN_STR_SHORT = 128
_MAX_LEN_PW = 128


def _validate_username(username: str) -> None:
    """
    Validate a username field in the request input.

    Parameters
    ----------
    username : str
        Username to validate.

    Raises
    ------
    ValidationError
        If the username is empty, too long or numerical
    """
    validate_name(username)
    username_regex = r"^[a-zA-Z][a-zA-Z0-9._-]+$"
    if re.match(username_regex, username) is None:
        raise ValidationError(
            f"Username {username} is invalid. Only letters, numbers, hyphens and "
            "underscores are allowed, and it should start with a letter."
        )


def _validate_password(password: str) -> None:
    """
    Check if the password is strong enough.

    Parameters
    ----------
    password : str
        Password to validate.

    Raises
    ------
    ValidationError
        If the password is not strong enough.
    """
    try:
        validate_password(password)
    except ValueError as e:
        raise ValidationError(str(e))


def _validate_organization_ids(organization_ids: list[int]) -> None:
    """
    Validate the organization ids in the input.

    Parameters
    ----------
    organization_ids : list[int]
        List of organization ids to validate.

    Raises
    ------
    ValidationError
        If the organization ids are not valid.
    """
    if not all(i > 0 for i in organization_ids):
        raise ValidationError("Organization ids must be greater than 0")
    if not len(organization_ids) == len(set(organization_ids)):
        raise ValidationError("Organization ids must be unique")
    if not len(organization_ids):
        raise ValidationError("At least one organization id is required")


class _OnlyIdSchema(Schema):
    """Schema for validating POST requests that only require an ID field."""

    id = fields.Integer(required=True, validate=Range(min=1))


class _NameValidationSchema(Schema):
    """Schema for validating POST requests with a name field."""

    name = fields.String(required=True)

    @validates("name")
    def validate_name(self, name: str):
        """
        Validate the name in the input.

        Parameters
        ----------
        name : str
            Name to validate.

        Raises
        ------
        ValidationError
            If the name is empty, too long or numerical
        """
        validate_name(name)


class _PasswordValidationSchema(Schema):
    """Schema that contains password validation function"""

    password = fields.String(required=True)

    @validates("password")
    def _validate_password(self, password: str):
        """
        Check if the password is strong enough.

        Parameters
        ----------
        password : str
            Password to validate.

        Raises
        ------
        ValidationError
            If the password is not strong enough.
        """
        _validate_password(password)


class ChangePasswordInputSchema(Schema):
    """Schema for validating input for changing a password."""

    # validation for current password is not necessary, as it is checked in the
    # authentication process
    current_password = fields.String(required=True, validate=Length(max=_MAX_LEN_PW))
    new_password = fields.String(required=True)

    @validates("new_password")
    def validate_password(self, password: str):
        """
        Check if the password is strong enough.

        Parameters
        ----------
        password : str
            Password to validate.

        Raises
        ------
        ValidationError
            If the password is not strong enough.
        """
        _validate_password(password)


class BasicAuthInputSchema(Schema):
    """Schema for validating input for basic authentication using a username and password."""

    username = fields.String(required=True, validate=Length(min=1, max=MAX_LEN_NAME))
    # Note that we don't inherit from _PasswordValidationSchema here and
    # don't validate password in case the password does not fulfill the
    # password policy. This is e.g. the case with the default root user created
    # when the server is started for the first time.
    password = fields.String(required=True, validate=Length(min=1, max=_MAX_LEN_PW))


class CollaborationInputSchema(_NameValidationSchema):
    """Schema for validating input for a creating a collaboration."""

    organization_ids = fields.List(fields.Integer(), required=True)
    encrypted = fields.Boolean(required=True)

    @validates("organization_ids")
    def validate_organization_ids(self, organization_ids):
        """
        Validate the organization ids in the input.

        Parameters
        ----------
        organization_ids : list[int]
            List of organization ids to validate.

        Raises
        ------
        ValidationError
            If the organization ids are not valid.
        """
        _validate_organization_ids(organization_ids)


class CollaborationChangeOrganizationSchema(_OnlyIdSchema):
    """
    Schema for validating requests that add an organization to a collaboration.
    """

    pass


class CollaborationAddNodeSchema(_OnlyIdSchema):
    """Schema for validating requests that add a node to a collaboration."""

    pass


class KillTaskInputSchema(_OnlyIdSchema):
    """Schema for validating input for killing a task."""

    pass


class KillNodeTasksInputSchema(_OnlyIdSchema):
    """Schema for validating input for killing tasks on a node."""

    pass


class NodeInputSchema(_NameValidationSchema):
    """Schema for validating input for a creating a node."""

    # overwrite name attr as it is not required for a node
    name = fields.String(required=False)
    collaboration_id = fields.Integer(required=True, validate=Range(min=1))
    organization_id = fields.Integer(validate=Range(min=1))
    ip = fields.String()
    clear_ip = fields.Boolean()

    @validates("ip")
    def validate_ip(self, ip: str):
        """
        Validate IP address in request body.

        Parameters
        ----------
        ip : str
            IP address to validate.

        Raises
        ------
        ValidationError
            If the IP address is not valid.
        """
        try:
            ipaddress.ip_address(ip)
        except ValueError:
            raise ValidationError("IP address is not valid")


class OrganizationInputSchema(_NameValidationSchema):
    """Schema for validating input for a creating an organization."""

    address1 = fields.String(validate=Length(max=_MAX_LEN_STR_SHORT))
    address2 = fields.String(validate=Length(max=_MAX_LEN_STR_SHORT))
    zipcode = fields.String(validate=Length(max=_MAX_LEN_STR_SHORT))
    country = fields.String(validate=Length(max=_MAX_LEN_STR_SHORT))
    domain = fields.String(validate=Length(max=_MAX_LEN_STR_SHORT))
    public_key = fields.String()


class PortInputSchema(Schema):
    """Schema for validating input for a creating a port."""

    port = fields.Integer(required=True)
    run_id = fields.Integer(required=True, validate=Range(min=1))
    label = fields.String(validate=Length(max=_MAX_LEN_STR_SHORT), allow_none=True)

    @validates("port")
    def validate_port(self, port):
        """
        Validate the port in the input.

        Parameters
        ----------
        port : int
            Port to validate.

        Raises
        ------
        ValidationError
            If the port is not valid.
        """
        if not 1 <= port <= 65535:
            raise ValidationError("Port must be between 1 and 65535")


class RecoverPasswordInputSchema(Schema):
    """Schema for validating input for recovering a password."""

    email = fields.Email()
    username = fields.String(validate=Length(max=MAX_LEN_NAME))

    @validates_schema
    def validate_email_or_username(self, data: dict, **kwargs) -> None:
        """
        Validate the input, which should contain either an email or username.

        Parameters
        ----------
        data : dict
            The input data. Should contain an email or username.

        Raises
        ------
        ValidationError
            If the input does not contain an email or username.
        """
        if not ("email" in data or "username" in data):
            raise ValidationError("Email or username is required")


class ResetPasswordInputSchema(_PasswordValidationSchema):
    """Schema for validating input for resetting a password."""

    reset_token = fields.String(required=True, validate=Length(max=MAX_LEN_STR_LONG))


class Recover2FAInputSchema(BasicAuthInputSchema):
    """Schema for validating input for recovering 2FA."""


class Reset2FAInputSchema(Schema):
    """Schema for validating input for resetting 2FA."""

    reset_token = fields.String(required=True, validate=Length(max=MAX_LEN_STR_LONG))


class ResetAPIKeyInputSchema(_OnlyIdSchema):
    """Schema for validating input for resetting an API key."""

    pass


class RunInputSchema(Schema):
    """Schema for validating input for patching an algorithm run."""

    started_at = fields.DateTime()
    finished_at = fields.DateTime()
    log = fields.String()
    result = fields.String()
    status = fields.String(validate=OneOf([s.value for s in TaskStatus]))


class TaskInputSchema(_NameValidationSchema):
    """Schema for validating input for creating a task."""

    # overwrite name attr as it is not required for a task
    name = fields.String(required=False)
    description = fields.String(validate=Length(max=MAX_LEN_STR_LONG))
    image = fields.String(required=True, validate=Length(min=1))
    collaboration_id = fields.Integer(validate=Range(min=1))
    study_id = fields.Integer(validate=Range(min=1))
    store_id = fields.Integer(validate=Range(min=1))
    server_url = fields.Url()
    organizations = fields.List(fields.Dict(), required=True)
    databases = fields.List(fields.Dict(), allow_none=True)

    @validates_schema
    def validate_collaboration_or_study(self, data: dict, **kwargs) -> None:
        """
        Validate the input, which should contain a collaboration_id or a study_id. The
        input may also contain both.

        Parameters
        ----------
        data : dict
            The input data.

        Raises
        ------
        ValidationError
            If the input does not contain a collaboration_id or a study_id.
        """
        if not ("collaboration_id" in data or "study_id" in data):
            raise ValidationError("Collaboration_id or study_id is required")

    @validates("organizations")
    def validate_organizations(self, organizations: list[dict]):
        """
        Validate the organizations in the input.

        Parameters
        ----------
        organizations : list[dict]
            List of organizations to validate. Each organization must have at
            least an organization id.

        Raises
        ------
        ValidationError
            If the organizations are not valid.
        """
        if not len(organizations):
            raise ValidationError("At least one organization is required")
        for organization in organizations:
            if "id" not in organization:
                raise ValidationError(
                    "Organization id is required for each organization"
                )

    @validates("databases")
    def validate_databases(self, databases: list[dict]):
        """
        Validate the databases in the input.

        Parameters
        ----------
        databases : list[dict]
            List of databases to validate. Each database must have at
            least a database label.

        Raises
        ------
        ValidationError
            If the databases are not valid.
        """
        if databases is None:
            return  # some algorithms don't use any database
        for database in databases:
            if "label" not in database:
                raise ValidationError("Database label is required for each database")
            if "preprocessing" in database:
                if not isinstance(database["preprocessing"], list):
                    raise ValidationError(
                        "Database preprocessing must be a list of dictionaries"
                    )
                # TODO we may add further validation on the preprocessing
                # parameters when that is completed
                for prepro in database["preprocessing"]:
                    if "function" not in prepro:
                        raise ValidationError(
                            f"Database preprocessing {prepro} is missing a "
                            "'function'"
                        )
            if "arguments" in database:
                if not isinstance(database["arguments"], dict):
                    raise ValidationError("Database arguments must be a dict")
            allowed_keys = {
                "label",
                "arguments",
                "preprocessing",
                "query",
                "sheet_name",
            }
            if not set(database.keys()).issubset(set(allowed_keys)):
                raise ValidationError(
                    f"Database {database} contains unknown keys. Allowed keys "
                    f"are {allowed_keys}."
                )


class TokenUserInputSchema(BasicAuthInputSchema):
    """Schema for validating input for creating a token for a user."""

    mfa_code = fields.String(validate=Length(max=10))

    # TODO in v5+, activate the code below to validate the username (allowed characters
    # etc). We cannot do this in v4 because some existing usernames do not comply and
    # those would then no longer be able to login
    # @validates("username")
    # def validate_username(self, username: str):
    #     """
    #     Check if the username is appropriate

    #     Parameters
    #     ----------
    #     username : str
    #         Username to validate.

    #     Raises
    #     ------
    #     ValidationError
    #         If the username is too short, too long or numeric.
    #     """
    #     _validate_username(username)


class TokenNodeInputSchema(Schema):
    """Schema for validating input for creating a token for a node."""

    api_key = fields.String(required=True)

    @validates("api_key")
    def validate_api_key(self, api_key: str):
        """
        Validate the API key in the input. The API key should be a valid UUID

        Parameters
        ----------
        api_key : str
            API key to validate.

        Raises
        ------
        ValidationError
            If the API key is not valid.
        """
        try:
            uuid.UUID(api_key)
        except ValueError:
            raise ValidationError("API key is not a valid UUID")


class TokenAlgorithmInputSchema(Schema):
    """Schema for validating input for creating a token for an algorithm."""

    task_id = fields.Integer(required=True, validate=Range(min=1))
    image = fields.String(required=True, validate=Length(min=1))


class UserInputSchema(_PasswordValidationSchema):
    """Schema for validating input for creating a user."""

    username = fields.String(required=True, validate=Length(min=3, max=MAX_LEN_NAME))
    email = fields.Email(required=True)
    firstname = fields.String(validate=Length(max=_MAX_LEN_STR_SHORT))
    lastname = fields.String(validate=Length(max=_MAX_LEN_STR_SHORT))
    organization_id = fields.Integer(validate=Range(min=1))
    roles = fields.List(fields.Integer(validate=Range(min=1)))
    rules = fields.List(fields.Integer(validate=Range(min=1)))

    @validates("username")
    def validate_username(self, username: str):
        """
        Check if the username is appropriate

        Parameters
        ----------
        username : str
            Username to validate.

        Raises
        ------
        ValidationError
            If the username is too short, too long or numeric.
        """
        _validate_username(username)


class VPNConfigUpdateInputSchema(Schema):
    """Schema for validating input for updating a VPN configuration."""

    vpn_config = fields.String(required=True)


class ColumnNameInputSchema(Schema):
    """Schema for validating input for collecting database column names."""

    db_label = fields.String(required=True)
    collaboration_id = fields.Integer(required=True, validate=Range(min=1))
    organizations = fields.List(fields.Dict(), required=True)
    sheet_name = fields.String(required=False)
    query = fields.String(required=False)

    @validates("organizations")
    def validate_organizations(self, organizations: list[dict]):
        """
        Validate the organizations in the input.

        Parameters
        ----------
        organizations : list[dict]
            List of organizations to validate. Each organization must have at
            least an organization id.

        Raises
        ------
        ValidationError
            If the organizations are not valid.
        """
        if not len(organizations):
            raise ValidationError("At least one organization is required")
        for organization in organizations:
            if "id" not in organization:
                raise ValidationError(
                    "Organization id is required for each organization"
                )


class AlgorithmStoreInputSchema(Schema):
    """Schema for validating input for creating an algorithm store."""

    name = fields.String(required=True)
    algorithm_store_url = fields.Url(required=True)
    server_url = fields.Url()
    collaboration_id = fields.Integer(validate=Range(min=1))
    force = fields.Boolean()


class StudyInputSchema(_NameValidationSchema):
    """Schema for validating input for creating a study"""

    collaboration_id = fields.Integer(required=True, validate=Range(min=1))
    organization_ids = fields.List(fields.Integer(), required=True)

    @validates("organization_ids")
    def validate_organization_ids(self, organization_ids):
        """
        Validate the organization ids in the input.

        Parameters
        ----------
        organization_ids : list[int]
            List of organization ids to validate.

        Raises
        ------
        ValidationError
            If the organization ids are not valid.
        """
        _validate_organization_ids(organization_ids)


class StudyChangeOrganizationSchema(_OnlyIdSchema):
    """
    Schema for validating requests that add an organization to a study.
    """

    pass



import logging

from flask import request, g
from flask_restful import Api
from http import HTTPStatus

from vantage6.server import db
from vantage6.backend.common.resource.pagination import Pagination
from vantage6.server.resource.common.input_schema import (
    CollaborationAddNodeSchema,
    CollaborationChangeOrganizationSchema,
    CollaborationInputSchema,
)
from vantage6.server.permission import (
    RuleCollection,
    Scope as S,
    Operation as P,
    PermissionManager,
)
from vantage6.server.resource.common.output_schema import (
    CollaborationSchema,
    CollaborationWithOrgsSchema,
    OrganizationSchema,
    NodeSchemaSimple,
)
from vantage6.server.resource import with_user, only_for, ServicesResources


module_name = __name__.split(".")[-1]
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the collaboration resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        Collaborations,
        path,
        endpoint="collaboration_without_id",
        methods=("GET", "POST"),
        resource_class_kwargs=services,
    )
    api.add_resource(
        Collaboration,
        path + "/<int:id>",
        endpoint="collaboration_with_id",
        methods=("GET", "PATCH", "DELETE"),
        resource_class_kwargs=services,
    )
    api.add_resource(
        CollaborationOrganization,
        path + "/<int:id>/organization",
        endpoint="collaboration_with_id_organization",
        methods=("POST", "DELETE"),
        resource_class_kwargs=services,
    )
    api.add_resource(
        CollaborationNode,
        path + "/<int:id>/node",
        endpoint="collaboration_with_id_node",
        methods=("POST", "DELETE"),
        resource_class_kwargs=services,
    )


# Schemas
collaboration_schema = CollaborationSchema()
org_schema = OrganizationSchema()
node_schema = NodeSchemaSimple()
collaboration_input_schema = CollaborationInputSchema()
collaboration_change_org_schema = CollaborationChangeOrganizationSchema()
collaboration_add_node_schema = CollaborationAddNodeSchema()
collab_with_orgs_schema = CollaborationWithOrgsSchema()


# -----------------------------------------------------------------------------
# Permissions
# -----------------------------------------------------------------------------
def permissions(permissions: PermissionManager) -> None:
    """
    Define the permissions for this resource.

    Parameters
    ----------
    permissions : PermissionManager
        Permission manager instance to which permissions are added
    """

    add = permissions.appender(module_name)

    add(scope=S.GLOBAL, operation=P.VIEW, description="view any collaboration")

    add(
        scope=S.ORGANIZATION,
        operation=P.VIEW,
        assign_to_container=True,
        assign_to_node=True,
        description="view collaborations of your organization",
    )

    add(scope=S.GLOBAL, operation=P.EDIT, description="edit any collaboration")
    add(
        scope=S.COLLABORATION,
        operation=P.EDIT,
        description="edit any collaboration that your organization " "participates in",
    )

    add(scope=S.GLOBAL, operation=P.CREATE, description="create a new collaboration")

    add(scope=S.GLOBAL, operation=P.DELETE, description="delete a collaboration")
    add(
        scope=S.COLLABORATION,
        operation=P.DELETE,
        description="delete any collaboration that your organization "
        "participates in",
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class CollaborationBase(ServicesResources):
    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        self.r: RuleCollection = getattr(self.permissions, module_name)

    def _select_schema(self) -> CollaborationSchema:
        """
        Select the output schema based on which resources should be included

        Returns
        -------
        CollaborationSchema or derivative of it
            Schema to use for serialization
        """
        if self.is_included("organizations"):
            return collab_with_orgs_schema
        else:
            return collaboration_schema


class Collaborations(CollaborationBase):
    @only_for(("user", "node"))
    def get(self):
        """Returns a list of collaborations
        ---
        description: >-
          Returns a list of collaborations. Depending on your permission, all
          collaborations are shown or only collaborations in which your
          organization participates. See the table bellow.\n\n

          ### Permission Table\n
          |Rulename|Scope|Operation|Assigned to Node|Assigned to Container|
          Description|\n
          | -- | -- | -- | -- | -- | -- |\n
          |Collaboration|Global|View|❌|❌|All collaborations|\n
          |Collaboration|Organization|View|✅|✅|Collaborations in which
          your organization participates |\n\n

          Accessible to users.

        parameters:
          - in: query
            name: name
            schema:
              type: string
            description: >-
              Name to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: encrypted
            schema:
              type: boolean
            description: Whether or not collaboration is encrypted
          - in: query
            name: organization_id
            schema:
              type: integer
            description: Organization id
          - in: query
            name: algorithm_store_id
            schema:
              type: integer
            description: Algorithm store ID that is available to the collaboration
          - in: query
            name: include
            schema:
              type: array
              items:
                type: string
            description: Include 'organizations' to include the organizations
              within the collaboration.
          - in: query
            name: page
            schema:
              type: integer
            description: Page number for pagination (default=1)
          - in: query
            name: per_page
            schema:
              type: integer
            description: Number of items per page (default=10)
          - in: query
            name: sort
            schema:
              type: string
            description: >-
              Sort by one or more fields, separated by a comma. Use a minus
              sign (-) in front of the field to sort in descending order.

        responses:
          200:
            description: Ok
          401:
            description: Unauthorized
          400:
            description: Improper values for pagination or sorting parameters

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """

        # obtain organization from authenticated
        auth_org_id = self.obtain_organization_id()
        q = g.session.query(db.Collaboration)
        args = request.args

        # filter by a field of this endpoint
        if "encrypted" in args:
            q = q.filter(db.Collaboration.encrypted == args["encrypted"])
        if "name" in args:
            q = q.filter(db.Collaboration.name.like(args["name"]))

        # find collaborations containing a specific organization
        if "organization_id" in args:
            if not self.r.v_glo.can() and args["organization_id"] != str(auth_org_id):
                return {
                    "msg": "You lack the permission to request collaborations "
                    "for this organization!"
                }, HTTPStatus.UNAUTHORIZED
            elif self.r.v_glo.can():
                q = (
                    q.join(db.Member)
                    .join(db.Organization)
                    .filter(db.Organization.id == args["organization_id"])
                )
            # else: no filter if user can only view collaborations of own
            # organization: the arg 'organization_id' is then superfluous

        if "algorithm_store_id" in args:
            # If this algorithm store is available for all collaborations, no filter is
            # needed
            store = db.AlgorithmStore.get(args["algorithm_store_id"])
            if not store:
                return {
                    "msg": f"Algorithm store with id={args['algorithm_store_id']} not found"
                }, HTTPStatus.NOT_FOUND
            elif store.collaboration_id is not None:
                # filter on collaborations that have access to this algorithm store
                q = q.join(db.AlgorithmStore).filter(
                    db.AlgorithmStore.collaboration_id == db.Collaboration.id,
                )

        # filter based on permissions
        if not self.r.v_glo.can():
            if self.r.v_org.can():
                q = q.join(db.Organization, db.Collaboration.organizations).filter(
                    db.Collaboration.organizations.any(id=auth_org_id)
                )
            else:
                return {
                    "msg": "You lack the permission to do that!"
                }, HTTPStatus.UNAUTHORIZED

        # paginate the results
        try:
            page = Pagination.from_query(q, request, db.Collaboration)
        except (ValueError, AttributeError) as e:
            return {"msg": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR

        schema = self._select_schema()

        # serialize models
        return self.response(page, schema)

    @with_user
    def post(self):
        """Create collaboration
        ---
        description: >-
          Create a new collaboration between organizations.\n\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to Node|Assigned to
          Container|Description|\n
          |--|--|--|--|--|--|\n
          |Collaboration|Global|Create|❌|❌|Create collaboration|\n\n

          Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  name:
                    type: string
                    description: Unique human readable name for collaboration
                  organization_ids:
                    type: array
                    items:
                      type: integer
                      description: List of organization ids which form the
                        collaboration
                  encrypted:
                    type: integer
                    description: Boolean (0 or 1) to indicate if the
                      collaboration uses encryption

        responses:
          200:
            description: Ok
          400:
            description: Collaboration name already exists
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """
        data = request.get_json()
        # validate request body
        errors = collaboration_input_schema.validate(data)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        name = data["name"]
        if db.Collaboration.exists("name", name):
            return {
                "msg": f"Collaboration name '{name}' already exists!"
            }, HTTPStatus.BAD_REQUEST

        if not self.r.c_glo.can():
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        encrypted = True if data["encrypted"] == 1 else False

        collaboration = db.Collaboration(
            name=name,
            organizations=[
                db.Organization.get(org_id)
                for org_id in data["organization_ids"]
                if db.Organization.get(org_id)
            ],
            encrypted=encrypted,
        )

        collaboration.save()
        return collaboration_schema.dump(collaboration), HTTPStatus.OK


class Collaboration(CollaborationBase):
    @only_for(("user", "node", "container"))
    def get(self, id):
        """Get collaboration
        ---
        description: >-
          Returns the collaboration with the specified id.\n

          ### Permission Table\n
          |Rulename|Scope|Operation|Assigned to Node|Assigned to Container|
          Description|\n
          | -- | -- | -- | -- | -- | -- |\n
          |Collaboration|Global|View|❌|❌|All collaborations|\n
          |Collaboration|Organization|View|✅|✅|Collaborations in which
          your organization participates |\n\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
              minimum: 1
            description: Collaboration id
            required: true
          - in: query
            name: include
            schema:
              type: array
              items:
                type: string
            description: Include 'organizations' to include the organizations
              within the collaboration.

        responses:
          200:
            description: Ok
          404:
            description: Collaboration with specified id is not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """
        collaboration = db.Collaboration.get(id)

        # check that collaboration exists
        if not collaboration:
            return {
                "msg": f"collaboration having id={id} not found"
            }, HTTPStatus.NOT_FOUND

        # obtain the organization id of the authenticated
        auth_org_id = self.obtain_organization_id()

        # verify that the user/node organization is within the
        # collaboration
        ids = [org.id for org in collaboration.organizations]
        if not self.r.v_glo.can():
            if not (self.r.v_org.can() and auth_org_id in ids):
                return {
                    "msg": "You lack the permission to do that!"
                }, HTTPStatus.UNAUTHORIZED

        schema = self._select_schema()

        return schema.dump(collaboration, many=False), HTTPStatus.OK  # 200

    @with_user
    def patch(self, id):
        """Update collaboration
        ---
        description: >-
          Updates the collaboration with the specified id.\n\n
          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Collaboration|Global|Edit|❌|❌|Update a collaboration|\n
          |Collaboration|Collaboration|Edit|❌|❌|Update a collaboration that
          you are already a member of|\n\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Collaboration id
            required: tr

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  name:
                    type: string
                    description: Human readable label
                  organization_ids:
                    type: array
                    items:
                      type: integer
                    description: List of organization ids
                  encrypted:
                    type: boolean
                    description: Whether collaboration is encrypted or not

        responses:
          200:
            description: Ok
          404:
            description: Collaboration with specified id is not found
          401:
            description: Unauthorized
          400:
            description: Collaboration name already exists

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """
        collaboration = db.Collaboration.get(id)

        # check if collaboration exists
        if not collaboration:
            return {
                "msg": f"collaboration having collaboration_id={id} " "can not be found"
            }, HTTPStatus.NOT_FOUND  # 404

        # verify permissions
        if not self.r.can_for_col(P.EDIT, collaboration.id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        data = request.get_json()
        # validate request body
        errors = collaboration_input_schema.validate(data, partial=True)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        # only update fields that are provided
        if "name" in data:
            name = data["name"]
            if collaboration.name != name and db.Collaboration.exists("name", name):
                return {
                    "msg": f"Collaboration name '{name}' already exists!"
                }, HTTPStatus.BAD_REQUEST
            collaboration.name = name
        if "organization_ids" in data:
            # Users with permission to edit a collaboration at collaboration scope are
            # not allowed to change the organizations of the collaboration: that would
            # be a permission escalation as doing so may allow them to also do other
            # actions for newly added organizations. Therefore, only users with global
            # permission can alter the organizations in a collaboration.
            if not self.r.e_glo.can():
                return {
                    "msg": "You lack the permission to change the organizations of a "
                    "collaboration!"
                }, HTTPStatus.UNAUTHORIZED
            # set new organizations
            collaboration.organizations = [
                db.Organization.get(org_id)
                for org_id in data["organization_ids"]
                if db.Organization.get(org_id)
            ]
        if "encrypted" in data:
            collaboration.encrypted = data["encrypted"]

        collaboration.save()

        return (
            collaboration_schema.dump(collaboration, many=False),
            HTTPStatus.OK,
        )  # 200

    @with_user
    def delete(self, id):
        """Delete collaboration
        ---
        description: >-
          Removes the collaboration from the database entirely.

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Collaboration|Global|Delete|❌|❌|Remove collaboration|\n
          |Collaboration|Collaboration|Delete|❌|❌|Remove collaborations
          that you are part of yourself|\n\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Collaboration id
            required: true
          - in: query
            name: delete_dependents
            schema:
              type: boolean
            description: If set to true, the collaboratio will be deleted along
              with all its tasks and nodes (default=False)

        responses:
          200:
            description: Ok
          404:
            description: Collaboration with specified id is not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """

        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {"msg": f"collaboration id={id} is not found"}, HTTPStatus.NOT_FOUND

        # verify permissions
        if not self.r.can_for_col(P.DELETE, collaboration.id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        if collaboration.tasks or collaboration.nodes or collaboration.studies:
            delete_dependents = request.args.get("delete_dependents", False)
            if not delete_dependents:
                return {
                    "msg": f"Collaboration id={id} has "
                    f"{len(collaboration.tasks)} tasks, {len(collaboration.nodes)} "
                    f"nodes and {len(collaboration.studies)} studies. Please delete "
                    "them separately or set delete_dependents=True"
                }, HTTPStatus.BAD_REQUEST
            else:
                log.warning(
                    "Deleting collaboration id=%s along with %s tasks, %s nodes and "
                    "% studies.",
                    id,
                    len(collaboration.tasks),
                    len(collaboration.nodes),
                    len(collaboration.studies),
                )
                for task in collaboration.tasks:
                    task.delete()
                for node in collaboration.nodes:
                    node.delete()
                for study in collaboration.studies:
                    study.delete()

        collaboration.delete()
        return {"msg": f"Collaboration id={id} successfully deleted"}, HTTPStatus.OK


class CollaborationOrganization(ServicesResources):
    """Resource for /api/collaboration/<int:id>/organization."""

    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        self.r: RuleCollection = getattr(self.permissions, module_name)

    @with_user
    def post(self, id):
        """Add organization to collaboration
        ---
        description: >-
          Adds a single organization to an existing collaboration.\n\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Collaboration|Global|Edit|❌|❌|Add organization to a
          collaboration|\n
          |Collaboration|Collaboration|Edit|❌|❌|Add organization to a
          collaboration that your organization is already a member of|\n\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Collaboration id
            required: tr

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  id:
                    type: integer
                    description: Organization id which needs to be added

        responses:
          200:
            description: Ok
          404:
            description: Specified collaboration or organization does not exist
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """
        # get collaboration to which te organization should be added
        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {
                "msg": f"collaboration having collaboration_id={id} can " "not be found"
            }, HTTPStatus.NOT_FOUND

        # verify permissions
        if not self.r.can_for_col(P.EDIT, collaboration.id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # validate request body
        data = request.get_json()
        errors = collaboration_change_org_schema.validate(data)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        # get the organization
        organization = db.Organization.get(data["id"])
        if not organization:
            return {
                "msg": f"organization with id={id} is not found"
            }, HTTPStatus.NOT_FOUND

        # append organization to the collaboration
        collaboration.organizations.append(organization)
        collaboration.save()
        return org_schema.dump(collaboration.organizations, many=True), HTTPStatus.OK

    @with_user
    def delete(self, id):
        """Remove organization from collaboration
        ---
        description: >-
          Removes a single organization from an existing collaboration.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Collaboration|Global|Edit|❌|❌|Remove an organization from an
          existing collaboration|\n
          |Collaboration|Collaboration|Edit|❌|❌|Remove an organization from
          an existing collaboration that your organization is a member of|\n\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Collaboration id
            required: true

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  id:
                    type: integer
                    description: Organization id which needs to be deleted

        responses:
          200:
            description: Ok
          404:
            description: Specified collaboration or organization does not exist
          401:
            description: Unauthorized

        tags: ["Collaboration"]
        """
        # get collaboration from which organization should be removed
        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {
                "msg": f"Collaboration with collaboration_id={id} can " "not be found"
            }, HTTPStatus.NOT_FOUND

        # validate requst body
        data = request.get_json()
        errors = collaboration_change_org_schema.validate(data)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        # get organization which should be deleted
        data = request.get_json()
        org_id = data["id"]
        organization = db.Organization.get(org_id)
        if not organization:
            return {
                "msg": f"Organization with id={org_id} is not found"
            }, HTTPStatus.NOT_FOUND

        if not self.r.can_for_col(P.EDIT, collaboration.id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # delete organization and update
        collaboration.organizations.remove(organization)
        collaboration.save()
        return org_schema.dump(collaboration.organizations, many=True), HTTPStatus.OK


class CollaborationNode(ServicesResources):
    """Resource for /api/collaboration/<int:id>/node."""

    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        self.r: RuleCollection = getattr(self.permissions, module_name)

    @with_user
    def post(self, id):
        """Add node to collaboration
        ---
        description: >-
          Add node to an existing collaboration.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Collaboration|Global|Edit|❌|❌|Add node to collaboration|\n
          |Collaboration|Collaboration|Edit|❌|❌|Add node to collaboration
          that your organization is a member of|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Collaboration id
            required: tr

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  id:
                    type: integer
                    description: ID of node to be added

        responses:
          201:
            description: Added node to collaboration
          404:
            description: Collaboration or node not found
          400:
            description: Node is already in collaboration
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """
        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {
                "msg": f"collaboration having collaboration_id={id} can " "not be found"
            }, HTTPStatus.NOT_FOUND

        if not self.r.can_for_col(P.EDIT, collaboration.id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # validate request body
        data = request.get_json()
        errors = collaboration_add_node_schema.validate(data)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        node = db.Node.get(data["id"])
        if not node:
            return {"msg": f"node id={data['id']} not found"}, HTTPStatus.NOT_FOUND

        if node in collaboration.nodes:
            return {
                "msg": f"node id={data['id']} is already in collaboration " f"id={id}"
            }, HTTPStatus.BAD_REQUEST
        elif node.organization not in collaboration.organizations:
            return {
                "msg": f"Node id={data['id']} belongs to an organization that "
                f"is not part of collaboration id={id}. Please add the "
                "organization to the collaboration first"
            }, HTTPStatus.BAD_REQUEST

        collaboration.nodes.append(node)
        collaboration.save()
        return node_schema.dump(collaboration.nodes, many=True), HTTPStatus.CREATED

    @with_user
    def delete(self, id):
        """Remove node from collaboration
        ---
        description: >-
          Removes a single node from an existing collaboration.

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Collaboration|Global|Edit|❌|❌|Remove node from collaboration|\n
          |Collaboration|Collaboration|Edit|❌|❌|Remove node from
          collaboration that your organization is a member of|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Collaboration id from which the node is to be deleted.
            required: true

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  id:
                    type: integer
                    description: Node id which needs to be deleted

        responses:
          200:
            description: Ok
          404:
            description: Collaboration or node not found
          400:
            description: Node is not part of the collaboration
          401:
            description: Unauthorized

        tags: ["Collaboration"]
        """
        collaboration = db.Collaboration.get(id)
        if not collaboration:
            return {
                "msg": f"collaboration having collaboration_id={id} can " "not be found"
            }, HTTPStatus.NOT_FOUND

        if not self.r.can_for_col(P.EDIT, collaboration.id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        data = request.get_json()
        node = db.Node.get(data["id"])
        if not node:
            return {"msg": f"node id={id} not found"}, HTTPStatus.NOT_FOUND

        if node not in collaboration.nodes:
            return {
                "msg": f"node id={data['id']} is not part of " f"collaboration id={id}"
            }, HTTPStatus.BAD_REQUEST

        collaboration.nodes.remove(node)
        collaboration.save()
        return {
            "msg": f"node id={data['id']} removed from collaboration " f"id={id}"
        }, HTTPStatus.OK



from __future__ import annotations
from sqlalchemy import Column, String, Integer, ForeignKey
from sqlalchemy.orm import relationship

from vantage6.server.model.base import Base, DatabaseSessionManager


class AlgorithmStore(Base):
    """
    Table that describes which algorithm stores are available for which
    collaborations.

    Attributes
    ----------
    name: str
        The name of the algorithm store
    url: str
        The url of the algorithm store
    collaboration_id: int
        The id of the
        :class:`~vantage6.server.model.collaboration.Collaboration` that this
        algorithm store belongs to. If it is ``None``, then it is available for
        all collaborations.
    """

    # fields
    name = Column(String)
    url = Column(String)
    collaboration_id = Column(Integer, ForeignKey("collaboration.id"))

    collaboration = relationship("Collaboration", back_populates="algorithm_stores")
    tasks = relationship("Task", back_populates="algorithm_store")

    @classmethod
    def get_by_url(cls, url: str) -> list[AlgorithmStore]:
        """
        Get all algorithm store records with a certain url

        Parameters
        ----------
        url : str
            The url of the algorithm store

        Returns
        -------
        list[AlgorithmStore]
            List of algorithm store records with that URL
        """
        session = DatabaseSessionManager.get_session()
        return session.query(AlgorithmStore).filter_by(url=url).all()

    def is_for_all_collaborations(self) -> bool:
        """
        Check if the algorithm store is available for all collaborations

        Returns
        -------
        bool
            True if the algorithm store is available for all collaborations,
            False otherwise
        """
        return self.collaboration_id is None



"""
The Member table is used to link organizations and collaborations together.
Each line in the table represents that a certain organization is member
of a certain collaboration by storing the ids of the organization and
collaboration.

The StudyMember table is used to link organizations and studies together. A study
is a subset of organizations that form a collaboration. The table is structured in the
same way as the Member table.
"""

from sqlalchemy import Column, Integer, ForeignKey, Table

from vantage6.server.model.base import Base


Member = Table(
    "Member",
    Base.metadata,
    Column("organization_id", Integer, ForeignKey("organization.id")),
    Column("collaboration_id", Integer, ForeignKey("collaboration.id")),
)

StudyMember = Table(
    "StudyMember",
    Base.metadata,
    Column("organization_id", Integer, ForeignKey("organization.id")),
    Column("study_id", Integer, ForeignKey("study.id")),
)



from enum import Enum

from vantage6.server.model.rule import Operation, Scope


class DefaultRole(str, Enum):
    """Enum containing the names of the default roles"""

    ROOT = "Root"
    CONTAINER = "container"
    NODE = "node"
    VIEWER = "Viewer"
    RESEARCHER = "Researcher"
    ORG_ADMIN = "Organization Admin"
    COL_ADMIN = "Collaboration Admin"

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


# TODO BvB 22-06-07: we now have to pass this 'db' module as argument to a
# function because that module has a connection to the database. This should
# not be necessary. Fix that after fixing the circular imports described in
# https://github.com/vantage6/vantage6/issues/53
def get_default_roles(db) -> list[dict]:
    """
    Get a list containing the default roles and their rules, so that they may
    be created in the database

    Parameters
    ----------
    db
        The vantage6.server.db module

    Returns
    -------
    List[Dict]
        A list with dictionaries that each describe one of the roles. Each role
        dictionary contains the following:

        name: str
            Name of the role
        description: str
            Description of the role
        rules: List[int]
            A list of rule id's that the role contains
    """
    # Define default roles
    # 1. Root user
    SUPER_ROLE = {
        "name": DefaultRole.ROOT,
        "description": "Super role",
        "rules": db.Rule.get(),
    }
    # 2. Role for viewing organization resources
    VIEWER_RULES = [
        db.Rule.get_by_("user", Scope.OWN, Operation.EDIT),
        db.Rule.get_by_("user", Scope.OWN, Operation.DELETE),
        db.Rule.get_by_("user", Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_("organization", Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_("organization", Scope.COLLABORATION, Operation.VIEW),
        db.Rule.get_by_("collaboration", Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_("role", Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_("node", Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_("node", Scope.COLLABORATION, Operation.VIEW),
        db.Rule.get_by_("task", Scope.COLLABORATION, Operation.VIEW),
        db.Rule.get_by_("run", Scope.COLLABORATION, Operation.VIEW),
        db.Rule.get_by_("port", Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_("event", Scope.ORGANIZATION, Operation.RECEIVE),
        db.Rule.get_by_("event", Scope.COLLABORATION, Operation.RECEIVE),
        db.Rule.get_by_("study", Scope.ORGANIZATION, Operation.VIEW),
    ]
    VIEWER_ROLE = {
        "name": DefaultRole.VIEWER,
        "description": "Can manage their own account and view resources "
        "related to their organization",
        "rules": VIEWER_RULES,
    }
    # 3. Researcher role
    RESEARCHER_RULES = VIEWER_RULES + [
        db.Rule.get_by_("task", Scope.COLLABORATION, Operation.CREATE),
        db.Rule.get_by_("task", Scope.ORGANIZATION, Operation.DELETE),
    ]
    RESEARCHER_ROLE = {
        "name": DefaultRole.RESEARCHER,
        "description": "Can perform tasks, manage their own account, and "
        "view resources related to their organization",
        "rules": RESEARCHER_RULES,
    }
    # 4. Organization administrator role
    ORG_ADMIN_RULES = VIEWER_RULES + [
        db.Rule.get_by_("user", Scope.ORGANIZATION, Operation.CREATE),
        db.Rule.get_by_("user", Scope.ORGANIZATION, Operation.EDIT),
        db.Rule.get_by_("user", Scope.ORGANIZATION, Operation.DELETE),
        db.Rule.get_by_("organization", Scope.ORGANIZATION, Operation.EDIT),
        db.Rule.get_by_("role", Scope.ORGANIZATION, Operation.CREATE),
        db.Rule.get_by_("role", Scope.ORGANIZATION, Operation.EDIT),
        db.Rule.get_by_("role", Scope.ORGANIZATION, Operation.DELETE),
        db.Rule.get_by_("node", Scope.ORGANIZATION, Operation.CREATE),
        db.Rule.get_by_("node", Scope.ORGANIZATION, Operation.EDIT),
        db.Rule.get_by_("event", Scope.ORGANIZATION, Operation.SEND),
    ]
    ORG_ADMIN_ROLE = {
        "name": DefaultRole.ORG_ADMIN,
        "description": "Can manage an organization including its users, roles, and nodes.",
        "rules": ORG_ADMIN_RULES,
    }
    # 5. Collaboration administrator role
    researcher_only_rules = [
        rule for rule in RESEARCHER_RULES if rule not in ORG_ADMIN_RULES
    ]
    COLLAB_ADMIN_RULES = ORG_ADMIN_RULES + researcher_only_rules + [
        db.Rule.get_by_("user", Scope.COLLABORATION, Operation.VIEW),
        db.Rule.get_by_("user", Scope.COLLABORATION, Operation.CREATE),
        db.Rule.get_by_("user", Scope.COLLABORATION, Operation.EDIT),
        # The following rule is given so that a collaboration admin can
        # view which organizations they may add to their collaboration
        db.Rule.get_by_("organization", Scope.GLOBAL, Operation.VIEW),
        db.Rule.get_by_("organization", Scope.COLLABORATION, Operation.EDIT),
        db.Rule.get_by_("collaboration", Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_("collaboration", Scope.COLLABORATION, Operation.EDIT),
        db.Rule.get_by_("role", Scope.COLLABORATION, Operation.VIEW),
        db.Rule.get_by_("node", Scope.COLLABORATION, Operation.CREATE),
        db.Rule.get_by_("node", Scope.COLLABORATION, Operation.VIEW),
        db.Rule.get_by_("node", Scope.COLLABORATION, Operation.DELETE),
        db.Rule.get_by_("event", Scope.COLLABORATION, Operation.SEND),
        db.Rule.get_by_("study", Scope.COLLABORATION, Operation.VIEW),
        db.Rule.get_by_("study", Scope.COLLABORATION, Operation.CREATE),
        db.Rule.get_by_("study", Scope.COLLABORATION, Operation.EDIT),
        db.Rule.get_by_("study", Scope.COLLABORATION, Operation.DELETE),
    ]
    COLLAB_ADMIN_ROLE = {
        "name": DefaultRole.COL_ADMIN,
        "description": "Can manage a collaboration including its organization and users."
        " Also has permissions of an organization admin and researcher.",
        "rules": COLLAB_ADMIN_RULES,
    }
    # Combine all in array
    return [SUPER_ROLE, VIEWER_ROLE, RESEARCHER_ROLE, ORG_ADMIN_ROLE, COLLAB_ADMIN_ROLE]



import base64
import hashlib
import logging
import os
import re
from http import HTTPStatus

import requests
from flask import request
from flask.globals import g
from flask_restful import Api
from requests_oauthlib import OAuth2Session

from vantage6.common import logger_name
from vantage6.server.exceptions import VPNConfigException, VPNPortalAuthException
from vantage6.server.resource import with_node, ServicesResources
from vantage6.server.resource.common.input_schema import VPNConfigUpdateInputSchema

module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the collaboration resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        VPNConfig,
        path,
        endpoint="vpn_config",
        methods=("GET",),
        resource_class_kwargs=services,
    )

    api.add_resource(
        VPNConfig,
        path + "/update",
        endpoint="vpn_config_update",
        methods=("POST",),
        resource_class_kwargs=services,
    )


vpn_config_schema = VPNConfigUpdateInputSchema()


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class VPNConfig(ServicesResources):
    @with_node
    def get(self):
        """Get OVPN configuration file
        ---
        description: >-
          Returns the contents of an OVPN configuration file. This
          configuration allows the node to connect to the VPN server.\n

          This endpoint is not accessible for users, but only for
          authenticated nodes.

        responses:
          200:
            description: Ok
          500:
            description: Error in server VPN configuration, or in authorizing
              to VPN portal to obtain VPN configuration file
          501:
            description: This server has no VPN service
          503:
            description: The VPN server cannot be reached

        security:
            - bearerAuth: []

        tags: ["VPN"]

        """
        # check if the VPN server is configured
        if not self._is_server_configured():
            return {
                "msg": "This server does not support VPN"
            }, HTTPStatus.NOT_IMPLEMENTED

        # obtain VPN config by calling EduVPN API
        try:
            vpn_connector = EduVPNConnector(self.config["vpn_server"])
            ovpn_config = vpn_connector.get_ovpn_config()
        except VPNPortalAuthException as e:
            log.error("Could not obtain VPN configuration file")
            log.error(e)
            return {
                "msg": (
                    "Could not obtain VPN configuration because the vantage6 "
                    "server could not authorize to the VPN portal. Please "
                    "contact your server administrator."
                )
            }, HTTPStatus.INTERNAL_SERVER_ERROR
        except VPNConfigException as e:
            log.error("Could not obtain VPN configuration file")
            log.error(e)
            return {
                "msg": (
                    "Could not obtain VPN configuration because the "
                    "vantage6 server is not properly configured for VPN. "
                    "Please contact your server administrator."
                )
            }, HTTPStatus.INTERNAL_SERVER_ERROR
        except requests.ConnectionError as e:
            log.critical(
                f"Node <{g.node.id}> tries to obtain a vpn config. "
                "However the VPN server is unreachable!"
            )
            log.exception(e)
            return {
                "msg": "VPN server unreachable. Please contact your server"
                " administrator"
            }, HTTPStatus.SERVICE_UNAVAILABLE

        return {"ovpn_config": ovpn_config}, HTTPStatus.OK

    def _is_server_configured(self) -> bool:
        """Check if vpn server is available in configuration"""
        if "vpn_server" not in self.config:
            log.warning(
                f"Node <{g.node.id}> tries to obtain a vpn config but "
                "this server has not configured a VPN server!"
            )
            return False
        return True


class EduVPNConnector:
    def __init__(self, vpn_config) -> None:
        """
        Provides API access to the VPN server

        Parameters
        ----------
        vpn_config : dict
            Contains the VPN server info
        """
        self.config = vpn_config
        self.check_config()
        endpoints = self._find_endpoints(self.config["url"])

        self.api_url = endpoints["api_endpoint"]
        self.authorization_url = endpoints["authorization_endpoint"]
        self.token_url = endpoints["token_endpoint"]

        # Base portal url is api url minus /api/v3
        self.portal_url = self.api_url[: -len("/api/v3")]
        self.authorization_response = None
        log.debug(
            f"Found eduvpn endpoints:\n"
            f"api: {self.api_url}\n"
            f"authorization: {self.authorization_url}\n"
            f"token: {self.token_url}\n"
        )

        log.debug(f"Portal url: {self.portal_url}")

        # The redirect url is never called, but needs to be the same as the url
        # set in the eduvpn server for this particular client.
        self.session = OAuth2Session(
            vpn_config["client_id"],
            scope="config",
            redirect_uri=vpn_config["redirect_url"],
        )

    def _find_endpoints(self, url: str) -> dict[str, str]:
        well_known_url = url + "/.well-known/vpn-user-portal"
        response = requests.get(well_known_url)

        well_known = response.json()
        v3_endpoints = well_known["api"]["http://eduvpn.org/api#3"]

        return v3_endpoints

    def check_config(self) -> None:
        """
        Check if any keys that have to be present in the VPN configuration
        are missing. Raises an error if keys are missing.
        """
        for key in [
            "url",
            "portal_username",
            "portal_userpass",
            "client_id",
            "client_secret",
            "redirect_url",
        ]:
            if not self.config.get(key):
                raise VPNConfigException(
                    f"The '{key}' parameter has not been defined in your "
                    "vpn_server settings. Please adjust your configuration "
                    "file."
                )

    def get_ovpn_config(self) -> str:
        """
        Obtain a configuration file from the VPN server. This file contains
        all details needed for a node to connect to the VPN server.

        Returns
        -------
        str (ovpn format)
            Open-vpn configuration file
        """

        # obtain access token if not set
        self._set_access_token()

        # get the user's profile id
        log.debug("Getting EduVPN profile information")
        profile = self.get_profile()
        profile_id = profile["info"]["profile_list"][0]["profile_id"]

        # get the OVPN configuration for the selected user profile
        log.debug("Obtaining OpenVPN configuration template")
        ovpn_config = self.connect(profile_id)

        return ovpn_config

    def _set_access_token(self) -> None:
        """Obtain an access token to enable access to EduVPN API"""
        if self.session.token:
            log.debug("EduVPN access token already acquired")
            return
        # set PKCE data (code challenge and code verifier)
        log.debug("Setting PKCE challenge")
        self._set_pkce()
        # call the authorization route of EduVPN to get authorization code
        log.debug("Authorizing to EduVPN portal")
        self._authorize()
        # use authorization code to obtain token
        log.debug("Obtaining token from EduVPN portal")
        self.session.token = self._get_token()

    def _set_pkce(self) -> None:
        """Generate PKCE code verifier and challenge"""
        # set PKCE verifier
        self.code_verifier = base64.urlsafe_b64encode(os.urandom(40)).decode("utf-8")
        self.code_verifier = re.sub("[^a-zA-Z0-9]+", "", self.code_verifier)

        # set PKCE challenge based on the verifier
        self.code_challenge = hashlib.sha256(
            self.code_verifier.encode("utf-8")
        ).digest()
        self.code_challenge = base64.urlsafe_b64encode(self.code_challenge).decode(
            "utf-8"
        )
        self.code_challenge = self.code_challenge.replace("=", "")

    def _authorize(self) -> None:
        """Call authorization route of EduVPN to get authorization code"""
        params = {
            "client_id": self.config["client_id"],
            "redirect_uri": self.config["redirect_url"],
            "response_type": "code",
            "scope": "config",
            "state": "some_string",
            "code_challenge_method": "S256",
            "code_challenge": self.code_challenge,
        }

        authorize_url, _ = self.session.authorization_url(
            self.authorization_url,
            code_challenge_method="S256",
            code_challenge=self.code_challenge,
        )
        log.debug(f"authorize_url: {authorize_url}")

        # Log in to eduvpn, receiving the authorization endpoint as redirect

        post_data = {
            "userName": self.config["portal_username"],
            "userPass": self.config["portal_userpass"],
            "_user_pass_auth_redirect_to": authorize_url,
            "authRedirectTo": authorize_url,
        }

        # Referer doesn't matter so much but is a required header
        headers = {
            "Referer": self.portal_url + "/home",
        }

        response = self.session.post(
            f"{self.portal_url}/_user_pass_auth/verify",
            data=post_data,
            headers=headers,
            allow_redirects=False,
        )

        # Now follow redirect to call authorization route, but prevent redirect. By preventing the
        # redirect, the authorization code can be used in the current session
        # which allows us to use the PKCE code verifier without having to save
        # it in e.g. database or session state.
        response = self.session.get(authorize_url, allow_redirects=False)

        log.debug(f"Authorization response headers: {response.headers}")
        log.debug(f"Authorization response{response}")
        self.authorization_response = response.headers["Location"]

        if not (200 <= response.status_code < 400):
            raise VPNPortalAuthException(
                "Authenticating to EduVPN failed. Please check the "
                "'client_id' and 'redirect_url' settings in your server "
                "configuration file."
            )
        elif "Location" not in response.headers:
            raise VPNPortalAuthException(
                "Authenticating to EduVPN failed. Please check the following "
                "settings of your configuration file: portal_username and "
                "portal_userpass."
            )

    def _get_token(self) -> dict:
        """
        Use authorization code to obtain a token from the EduVPN portal. It seems that the
        vpn client can either be considered a confidential or a public client.

        Returns
        -------
        Dict:
            EduVPN portal token
        """

        errors = []
        token = None
        for include_client_id in [True, False]:
            try:
                token = self.session.fetch_token(
                    self.token_url,
                    authorization_response=self.authorization_response,
                    client_id=self.config["client_id"],
                    client_secret=self.config["client_secret"],
                    code_verifier=self.code_verifier,
                    include_client_id=include_client_id,
                )
                break
            except Exception as e:
                errors.append(e)
                continue

        if token is None:
            raise Exception(f"Failed to fetch token. Errors: {errors}")

        return token

    def get_profile(self) -> dict:
        """
        Call the profile_list route of EduVPN API

        Returns
        -------
        Dict
            Response content of the EduVPN /profile_list route
        """
        response = self.session.get(f"{self.api_url}/info")
        return response.json()

    def connect(self, profile_id: str) -> str:
        """Retrieve a vpn certificate

        Parameters
        ----------
        profile_id: str
            An EduVPN user's profile_id obtained from the /info` route

        Returns
        -------
        str
            OpenVPN configuration file content (without key pair)
        """
        params = {"profile_id": profile_id}
        response = self.session.post(self.api_url + "/connect", data=params)
        return response.content.decode("utf-8")



import datetime as dt
import logging

from functools import wraps

from flask import g, request
from flask_restful import Api
from flask_mail import Mail
from flask_jwt_extended import get_jwt, get_jwt_identity, jwt_required
from flask_socketio import SocketIO


from vantage6.common import logger_name
from vantage6.backend.common.services_resources import BaseServicesResources
from vantage6.server import db
from vantage6.server.model.authenticatable import Authenticatable
from vantage6.server.permission import (
    PermissionManager,
    obtain_auth_collaborations,
    obtain_auth_organization,
)
from vantage6.server.service.azure_storage_service import AzureStorageService

log = logging.getLogger(logger_name(__name__))


class ServicesResources(BaseServicesResources):
    """
    Flask resource class for the vantage6 server.

    Adds functionality like mail, socket, permissions and the api itself.
    Also adds common helper functions.

    Attributes
    ----------
    socketio : SocketIO
        SocketIO instance
    storage_adapter : AzureStorageService | None
        Storage adapter for handling large files
    mail : Mail
        Mail instance
    api : Api
        Api instance
    permissions : PermissionManager
        Instance of class that manages permissions
    config : dict
        Configuration dictionary
    """

    def __init__(
        self,
        socketio: SocketIO,
        storage_adapter: AzureStorageService | None,
        mail: Mail,
        api: Api,
        permissions: PermissionManager,
        config: dict,
    ):
        super().__init__(api, config, permissions, mail)
        self.socketio = socketio
        self.storage_adapter = storage_adapter

    @staticmethod
    def obtain_auth() -> db.Authenticatable | dict:
        """
        Read authenticatable object or dict from the flask global context.

        Returns
        -------
        Union[db.Authenticatable, dict]
            Authenticatable object or dict. Authenticatable object is either a
            user or node. Dict is for a container.
        """
        if g.user:
            return g.user
        if g.node:
            return g.node
        if g.container:
            return g.container

    @staticmethod
    def obtain_organization_id() -> int:
        """
        Obtain the organization id from the auth that is logged in.

        Returns
        -------
        int
            Organization id
        """
        if g.user:
            return g.user.organization.id
        elif g.node:
            return g.node.organization.id
        else:
            return g.container["organization_id"]

    @classmethod
    def obtain_auth_organization(cls) -> db.Organization:
        """
        Obtain the organization model from the auth that is logged in.

        Returns
        -------
        db.Organization
            Organization model
        """
        return obtain_auth_organization()

    @staticmethod
    def obtain_auth_collaborations() -> list[db.Collaboration]:
        """
        Obtain the collaborations that the auth is part of.

        Returns
        -------
        list[db.Collaboration]
            List of collaborations
        """
        return obtain_auth_collaborations()

    @staticmethod
    def obtain_auth_collaboration_ids() -> list[int]:
        """
        Obtain the collaboration ids that the auth is part of.

        Returns
        -------
        list[int]
            List of collaboration ids
        """
        return [col.id for col in obtain_auth_collaborations()]


# ------------------------------------------------------------------------------
# Helper functions/decoraters ...
# ------------------------------------------------------------------------------
def only_for(types: tuple[str] = ("user", "node", "container")) -> callable:
    """
    JWT endpoint protection decorator

    Parameters
    ----------
    types : list[str]
        List of types that are allowed to access the endpoint. Possible types
        are 'user', 'node' and 'container'.

    Returns
    -------
    function
        Decorator function that can be used to protect endpoints
    """

    def protection_decorator(fn):
        @wraps(fn)
        def decorator(*args, **kwargs):
            # decode JWT-token
            identity = get_jwt_identity()
            claims = get_jwt()

            # check that identity has access to endpoint
            g.type = claims["client_type"]

            if g.type not in types:
                # FIXME BvB 23-10-19: user gets a 500 error, would be better to
                # get an error message with 400 code
                msg = (
                    f"{g.type}s are not allowed to access {request.url} "
                    f"({request.method})"
                )
                log.warning(msg)
                raise Exception(msg)

            # do some specific stuff per identity
            g.user = g.container = g.node = None

            if g.type == "user":
                user = get_and_update_authenticatable_info(identity)
                g.user = user
                assert g.user.type == g.type
                log.debug(f"Received request from user {user.username} ({user.id})")

            elif g.type == "node":
                node = get_and_update_authenticatable_info(identity)
                g.node = node
                assert g.node.type == g.type
                log.debug(f"Received request from node {node.name} ({node.id})")

            elif g.type == "container":
                g.container = identity
                log.debug(
                    "Received request from container with node id "
                    f"{identity['node_id']} and task id {identity['task_id']}"
                )

            else:
                raise Exception(f"Unknown entity: {g.type}")

            return fn(*args, **kwargs)

        return jwt_required()(decorator)

    return protection_decorator


def get_and_update_authenticatable_info(auth_id: int) -> db.Authenticatable:
    """
    Get user or node from ID and update last time seen online.

    Parameters
    ----------
    auth_id : int
        ID of the user or node

    Returns
    -------
    db.Authenticatable
        User or node database model
    """
    auth = db.Authenticatable.get(auth_id)
    auth.last_seen = dt.datetime.now(dt.timezone.utc)
    auth.save()
    return auth


# create alias decorators
with_user_or_node = only_for(
    (
        "user",
        "node",
    )
)
with_user = only_for(("user",))
with_node = only_for(("node",))
with_container = only_for(("container",))


def get_org_ids_from_collabs(auth: Authenticatable, collab_id: int = None) -> list[int]:
    """
    Get all organization ids from the collaborations the user or node is in.

    Parameters
    ----------
    auth : Authenticatable
        User or node
    collab_id : int, optional
        Collaboration id. If given, only return the organization ids of this
        collaboration. If not given, return all organization ids of all
        collaborations the user or node is in.

    Returns
    -------
    list[int]
        List of organization ids
    """
    if collab_id:
        return [
            org.id
            for col in auth.organization.collaborations
            for org in col.organizations
            if col.id == collab_id
        ]
    else:
        return [
            org.id
            for col in auth.organization.collaborations
            for org in col.organizations
        ]



import logging

from http import HTTPStatus
from flask import request, g
from flask_restful import Api
from sqlalchemy import or_

from vantage6.server import db
from vantage6.backend.common.resource.pagination import Pagination
from vantage6.backend.common import get_server_url
from vantage6.server.resource.common.input_schema import AlgorithmStoreInputSchema
from vantage6.server.permission import RuleCollection, Operation as P
from vantage6.server.resource.common.output_schema import AlgorithmStoreSchema
from vantage6.server.resource import with_user, with_user_or_node, ServicesResources
from vantage6.server.algo_store_communication import (
    post_algorithm_store,
    request_algo_store,
)


module_name = __name__.split(".")[-1]
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the collaboration resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info('Setting up "%s" and subdirectories', path)

    api.add_resource(
        AlgorithmStores,
        api_base + "/algorithmstore",
        endpoint="algorithm_store_without_id",
        methods=("GET", "POST"),
        resource_class_kwargs=services,
    )
    api.add_resource(
        AlgorithmStore,
        api_base + "/algorithmstore/<int:id>",
        endpoint="algorithm_store_with_id",
        methods=("GET", "PATCH", "DELETE"),
        resource_class_kwargs=services,
    )


algorithm_store_schema = AlgorithmStoreSchema()
algorithm_store_input_schema = AlgorithmStoreInputSchema()


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class AlgorithmStoreBase(ServicesResources):
    """Base class for algorithm store resources."""

    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        self.r_col: RuleCollection = getattr(self.permissions, "collaboration")

    @staticmethod
    def get_authorization_headers_from_request() -> dict:
        """
        Get the authorization headers from the request.

        Returns
        -------
        dict
            The authorization headers
        """
        current_headers = request.headers
        if "Authorization" in current_headers:
            return {"Authorization": current_headers["Authorization"]}
        return {}


class AlgorithmStores(AlgorithmStoreBase):
    """Resource for /algorithm"""

    @with_user
    def get(self):
        """Returns a list of available algorithm stores
        ---
        description: >-
          Returns a list of available algorithm stores. Depending on your
          permission, all algorithm stores are shown or only algorithm stores
          in which your organization participates.\n\n

          ### Permission Table\n
          |Rulename|Scope|Operation|Assigned to Node|Assigned to Container|
          Description|\n
          | -- | -- | -- | -- | -- | -- |\n
          |Collaboration|Global|View|❌|❌|All algorithm stores|\n
          |Collaboration|Organization|View|❌|❌|Algorithm stores within
          collaborations in which your organization participates|\n\n

          Accessible to users.

        parameters:
          - in: query
            name: name
            schema:
              type: string
            description: >-
              Name to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: url
            schema:
              type: string
            description: >-
              Algorithm store URL to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: collaboration_id
            schema:
              type: integer
            description: Collaboration id
          - in: query
            name: page
            schema:
              type: integer
            description: Page number for pagination (default=1)
          - in: query
            name: per_page
            schema:
              type: integer
            description: Number of items per page (default=10)
          - in: query
            name: sort
            schema:
              type: string
            description: >-
              Sort by one or more fields, separated by a comma. Use a minus
              sign (-) in front of the field to sort in descending order.

        responses:
          200:
            description: Ok
          401:
            description: Unauthorized
          400:
            description: Improper values for pagination or sorting parameters
          404:
            description: Collaboration with specified id is not found

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """
        # TODO make an option to GET algorithm stores that are available to all
        # collaborations

        # obtain organization from authenticated
        auth_org = self.obtain_auth_organization()
        q = g.session.query(db.AlgorithmStore)
        args = request.args

        # filter by a field of this endpoint
        for param in ["name", "url"]:
            if param in args:
                q = q.filter(getattr(db.AlgorithmStore, param).like(args[param]))

        if "collaboration_id" in args:
            collab = db.Collaboration.get(args["collaboration_id"])
            if not collab:
                return {"msg": "Collaboration not found"}, HTTPStatus.NOT_FOUND
            # TODO generalize this check (make it available as part of the
            # permission class)
            if not self.r_col.v_glo.can():
                orgs_in_collab = [org.id for org in collab.organizations]
                if not (self.r_col.v_org.can() and auth_org.id in orgs_in_collab):
                    return {
                        "msg": "You lack the permission to do that!"
                    }, HTTPStatus.UNAUTHORIZED
            q = q.filter(
                or_(
                    db.AlgorithmStore.collaboration_id == args["collaboration_id"],
                    db.AlgorithmStore.collaboration_id.is_(None),
                )
            )

        # filter based on permissions
        if not self.r_col.v_glo.can():
            collab_ids = [c.id for c in auth_org.collaborations]
            if self.r_col.v_org.can():
                q = q.filter(
                    or_(
                        db.AlgorithmStore.collaboration_id.in_(collab_ids),
                        db.AlgorithmStore.collaboration_id.is_(None),
                    )
                )
            else:
                return {
                    "msg": "You lack the permission to do that!"
                }, HTTPStatus.UNAUTHORIZED

        # paginate the results
        try:
            page = Pagination.from_query(q, request, db.AlgorithmStore)
        except (ValueError, AttributeError) as e:
            return {"msg": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR

        # serialize models
        return self.response(page, algorithm_store_schema)

    @with_user
    def post(self):
        """Add algorithm store to a collaboration
        ---
        description: >-
          Register an algorithm store to a collaboration. Its algorithms will
          be available for that collaboration from then on.\n\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to Node|Assigned to
          Container|Description|\n
          |--|--|--|--|--|--|\n
          |Collaboration|Global|Edit|❌|❌|Add algorithm store to a
          collaboration|\n
          |Collaboration|Collaboration|Edit|❌|❌|Add algorithm store to a
          collaboration that your organization is a member of|\n\n

          Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  name:
                    type: string
                    description: Human readable name for the algorithm store
                  algorithm_store_url:
                    type: string
                    description: URL to the algorithm store, including the API path
                  server_url:
                    type: string
                    description: URL to this vantage6 server. This is used to
                      whitelist this server at the algorithm store. Note that
                      this is ignored if the server configuration contains
                      the 'server_url' key.
                  collaboration_id:
                    type: integer
                    description: Collaboration id to which the algorithm store
                      will be added. If not given, the algorithm store will be
                      available to all collaborations.
                  force:
                    type: boolean
                    description: Force adding the algorithm store to the
                      collaboration. This will overwrite warnings if insecure
                      addresses (e.g. localhost) are added.

        responses:
          200:
            description: Ok
          400:
            description: Wrong input data, algorithm store is already
              available for the collaboration or algorithm store is unreachable
          401:
            description: Unauthorized
          404:
            description: Collaboration with specified id is not found

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """
        data = request.get_json()
        # validate request body
        errors = algorithm_store_input_schema.validate(data)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        # check if collaboration exists
        collaboration_id = data.get("collaboration_id", None)
        if collaboration_id:
            collaboration_id = int(collaboration_id)
            collaboration = db.Collaboration.get(collaboration_id)
            if not collaboration:
                return {"msg": "Collaboration not found"}, HTTPStatus.NOT_FOUND

        # check permissions
        if not collaboration_id and not self.r_col.e_glo.can():
            return {
                "msg": "You lack the permission to add algorithm stores "
                "for all collaborations!"
            }, HTTPStatus.UNAUTHORIZED
        elif not self.r_col.can_for_col(P.EDIT, collaboration_id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        response, status = post_algorithm_store(
            request.get_json(),
            self.config,
            headers=self.get_authorization_headers_from_request(),
        )
        if status != HTTPStatus.CREATED:
            return response, status
        else:
            return algorithm_store_schema.dump(response), status


class AlgorithmStore(AlgorithmStoreBase):
    """Resource for /algorithm/<id>"""

    @with_user_or_node
    def get(self, id):
        """Get algorithm store record
        ---
        description: >-
          Returns a specific algorithm store record.\n

          ### Permission Table\n
          |Rulename|Scope|Operation|Assigned to Node|Assigned to Container|
          Description|\n
          | -- | -- | -- | -- | -- | -- |\n
          |Collaboration|Global|View|❌|❌|All algorithm stores|\n
          |Collaboration|Organization|View|❌|❌|Algorithm stores of
          collaborations in which your organization participates|\n\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
              minimum: 1
            description: Collaboration id
            required: true

        responses:
          200:
            description: Ok
          404:
            description: Algorithm store with specified id is not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """
        algorithm_store = db.AlgorithmStore.get(id)

        # check that collaboration exists, unlikely to happen without ID
        if not algorithm_store:
            return {"msg": f"Algorithm store id={id} not found"}, HTTPStatus.NOT_FOUND

        # verify that the user organization is within the collaboration
        # algorithm stores that are available to all collaborations are
        # always visible
        if not self.r_col.v_glo.can():
            auth_org_id = self.obtain_organization_id()
            ids = []
            if algorithm_store.collaboration:
                ids = [org.id for org in algorithm_store.collaboration.organizations]
            if not (
                self.r_col.v_org.can()
                and (auth_org_id in ids or not algorithm_store.collaboration_id)
            ):
                return {
                    "msg": "You lack the permission to do that!"
                }, HTTPStatus.UNAUTHORIZED

        return (
            algorithm_store_schema.dump(algorithm_store, many=False),
            HTTPStatus.OK,
        )  # 200

    # Note that this endpoint cannot be used to change the collaboration_id
    # or the url of the algorithm store, because those actions make little
    # sense and/or require additional checks. Users should then delete the
    # algorithm store link and create a new one.
    @with_user
    def patch(self, id):
        """Update algorithm store record

        Note that the algorithm store's URL cannot be changed, because it cannot be
        checked if the new URL represents the same algorithm store as the old URL. In
        such cases, please delete and re-add the algorithm store link.

        ---
        description: >-
          Updates the linked algorithm store with the specified id. Note that
          you cannot change the collaboration or the algorithm store URL: these
          can only be modified by deleting and re-creating the algorithm store
          link.\n\n
          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Collaboration|Global|Edit|❌|❌|Update any algorithm store|\n
          |Collaboration|Collaboration|Edit|❌|❌|Update algorithm stores
          within a collaboration that your organization is a member of|\n\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Algorithm store id
            required: tr

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  name:
                    type: string
                    description: Human readable label
                  collaboration_id:
                    type: integer
                    description: Collaboration id to which the algorithm store
                      will be added. If set to None, the algorithm store will
                      be available to all collaborations.

        responses:
          200:
            description: Ok
          400:
            description: Wrong input data
          404:
            description: Algorithm store with specified id is not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """
        algorithm_store = db.AlgorithmStore.get(id)
        # check if collaboration exists
        if not algorithm_store:
            return {
                "msg": f"Algorithm store with id={id} can not be found"
            }, HTTPStatus.NOT_FOUND  # 404

        # validate request body
        data = request.get_json()
        errors = algorithm_store_input_schema.validate(data, partial=True)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        # verify permissions - check permission for old collaboration
        collaboration_id_old = algorithm_store.collaboration_id
        if not self.r_col.can_for_col(P.EDIT, collaboration_id_old):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # verify permissions - check permission for new collaboration (if
        # specified) AND the old one
        data = request.get_json()
        if "collaboration_id" in data:
            collaboration_id_new = data["collaboration_id"]
            if not self.r_col.can_for_col(
                P.EDIT, collaboration_id_new
            ) or not self.r_col.can_for_col(P.EDIT, collaboration_id_old):
                return {
                    "msg": "You lack the permission to do that!"
                }, HTTPStatus.UNAUTHORIZED

        # only update fields that are provided
        name = data.get("name")
        if name is not None:
            algorithm_store.name = name

        # update collaboration_id if specified - also if it is set to None (
        # that makes it available to all collaborations)
        if "collaboration_id" in data:
            algorithm_store.collaboration_id = data["collaboration_id"]

        algorithm_store.save()

        return (
            algorithm_store_schema.dump(algorithm_store, many=False),
            HTTPStatus.OK,
        )  # 200

    @with_user
    def delete(self, id):
        """Delete linked algorithm store record
        ---
        description: >-
          Removes the algorithm store from the database entirely.

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Collaboration|Global|Edit|❌|❌|Remove any algorithm store from any
          collaboration|\n
          |Collaboration|Collaboration|Edit|❌|❌|Remove any algorithm store
          from a collaboration that your organization is a member of|\n\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Algorithm store id
            required: true
          - in: query
            name: server_url
            schema:
              type: string
            description: URL to this vantage6 server. This is used to delete
              the whitelisting of this server at the algorithm store. Note that
              this is ignored if the server configuration contains the
              'server_url' key.

        responses:
          200:
            description: Ok
          404:
            description: Algorithm store with specified id is not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """

        algorithm_store = db.AlgorithmStore.get(id)
        if not algorithm_store:
            return {
                "msg": f"Algorithm store id={id} is not found"
            }, HTTPStatus.NOT_FOUND

        # verify permissions
        if not self.r_col.can_for_col(P.EDIT, algorithm_store.collaboration_id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # Check if algostore is used by other collaborations. If it is, user needs
        # to have global permissions to delete all those links.
        # TODO require extra --force parameter to do this?
        all_stores_with_url = db.AlgorithmStore.get_by_url(algorithm_store.url)
        if len(all_stores_with_url) > 1:
            if self.r_col.e_glo.can():
                log.warning(
                    "Deleting algorithm store with id=%s, which is also used by"
                    " other collaborations",
                    id,
                )
            else:
                return {
                    "msg": "This algorithm store is used by other collaborations. "
                    "You lack the permission to delete it!"
                }, HTTPStatus.UNAUTHORIZED

        # only this algorithm store uses this url, so delete the
        # whitelisting
        server_url = get_server_url(self.config, request.args.get("server_url"))
        if not server_url:
            return {
                "msg": "The 'server_url' query parameter is required"
            }, HTTPStatus.BAD_REQUEST

        # get the ID of the whitelisted server, then delete it
        response, status = request_algo_store(
            algorithm_store.url,
            server_url,
            endpoint="vantage6-server",
            method="get",
            headers=self.get_authorization_headers_from_request(),
        )
        if status == HTTPStatus.FORBIDDEN:
            log.info(
                "Server with url=%s was not whitelisted at the algorithm store. "
                "Proceeding to remove algorithm store store from server...",
                server_url,
            )
            # initialize empty list of servers to delete at store
            result = []
        elif status != HTTPStatus.OK:
            return response, status
        else:
            result = response.json()
            if len(result) > 1:
                msg = (
                    "More than one whitelisted server found with url "
                    f"{server_url}. This should not happen! All will be "
                    "removed."
                )
                log.warning(msg)
            elif len(result) == 0:
                msg = (
                    "No whitelisted server found with url "
                    f"{server_url}. This should not happen!"
                )
                log.warning(msg)
        # remove all linked servers with the given url
        for server in result:
            server_id = server["id"]
            response, status = request_algo_store(
                algorithm_store.url,
                server_url,
                endpoint=f"vantage6-server/{server_id}",
                method="delete",
                headers=self.get_authorization_headers_from_request(),
            )
            if status != HTTPStatus.OK:
                return response, status

        # remove the store link from all tasks linked to this store
        for task in algorithm_store.tasks:
            task.store = None
            task.save()

        # finally delete the algorithm store record itself
        # pylint: disable=expression-not-assigned
        [store.delete() for store in all_stores_with_url]
        return {"msg": f"Algorithm store id={id} successfully deleted"}, HTTPStatus.OK



import sys
import logging
import datetime as dt
import pyotp

from http import HTTPStatus
from flask import request, render_template, current_app, Flask
from flask_mail import Mail
from threading import Thread

from vantage6.common.globals import APPNAME
from vantage6.backend.common.globals import (
    DEFAULT_EMAIL_FROM_ADDRESS,
    DEFAULT_SUPPORT_EMAIL_ADDRESS,
)
from vantage6.server.globals import (
    DEFAULT_MAX_FAILED_ATTEMPTS,
    DEFAULT_INACTIVATION_MINUTES,
    DEFAULT_BETWEEN_USER_EMAILS_MINUTES,
)
from vantage6.server.model.user import User

module_name = __name__.split(".")[-1]
log = logging.getLogger(module_name)


def user_login(
    config: dict, username: str, password: str, mail: Mail
) -> tuple[dict | User, HTTPStatus]:
    """
    Returns user a message in case of failed login attempt.

    config: dict
        Dictionary with configuration settings
    username: str
        Username of user to be logged in
    password: str
        Password of user to be logged in
    mail: flask_mail.Mail
        An instance of the Flask mail class. Used to send email to user in case
        of too many failed login attempts.

    Returns
    -------
    :class:`~vantage6.server.model.user.User` or dict:
        User SQLAlchemy model if user is logged in, otherwise dictionary with
        error message
    HTTPStatus:
        Status code that the current request should return
    """
    log.info("Trying to login '%s'", username)
    failed_login_msg = "Failed to login"

    # check if username exists. If it does not, we continue anyway, to prevent
    # that an attacker can find out which usernames exist via a timing attack.
    # In that case, we fetch the first user as random user.
    username_exists = User.username_exists(username)
    random_username = User.get_first_user().username
    user = (
        User.get_by_username(username)
        if username_exists
        else User.get_by_username(random_username)
    )

    password_policy = config.get("password_policy", {})
    max_failed_attempts = password_policy.get(
        "max_failed_attempts", DEFAULT_MAX_FAILED_ATTEMPTS
    )
    inactivation_time = password_policy.get(
        "inactivation_minutes", DEFAULT_INACTIVATION_MINUTES
    )

    is_blocked, min_rem = user.is_blocked(max_failed_attempts, inactivation_time)

    if user.check_password(password) and not is_blocked and username_exists:
        # Note: above the username_exists is checked to prevent that an
        # attacker happens to get the correct password for the random user
        # that is returned when the username does not exist. Note also that
        # the password is checked first to keep the timing equal for both.
        user.failed_login_attempts = 0
        user.save()
        return user, HTTPStatus.OK

    # Handle database updates required upon failed login in a separate thread
    # to ensure similar response times
    # pylint: disable=W0212
    t1 = Thread(
        target=__handle_failed_login,
        args=(
            current_app._get_current_object(),
            username_exists,
            username,
            password_policy,
            is_blocked,
            min_rem,
            mail,
            config,
            request.access_route[-1],
        ),
    )
    t1.start()

    return {"msg": failed_login_msg}, HTTPStatus.UNAUTHORIZED


def __handle_failed_login(
    app: Flask,
    user_exists: bool,
    username: str,
    password_policy: dict,
    is_blocked: bool,
    min_rem: int,
    mail: Mail,
    config: dict,
    ip: str,
) -> None:
    """
    When a user login fails, this function is called to update the database
    with the failed login attempt and send an email to the user if necessary.

    Note that this function is called in a separate thread to keep response
    times for login attempts similar in all cases. Therefore, this function
    calls `sys.exit()` to terminate the thread.

    Parameters
    ----------
    app: flask.Flask
        The current Flask app
    user_exists: bool
        Whether user exists or not
    username: str
        Username of the user that failed to login
    password_policy: dict
        Dictionary with password policy settings.
    min_rem: int
        Number of minutes remaining before the account is unlocked
    mail: flask_mail.Mail
        An instance of the Flask mail class. Used to send email to user in case
        of too many failed login attempts.
    config: dict
        Dictionary with configuration settings
    ip: str
        IP address from where the login attempt was made
    """
    if not user_exists:
        sys.exit()
    # get user object again (required because we are in a new thread)
    user = User.get_by_username(username)

    max_failed_attempts = password_policy.get(
        "max_failed_attempts", DEFAULT_MAX_FAILED_ATTEMPTS
    )

    if is_blocked:
        # alert the user via email that they are blocked
        __notify_user_blocked(app, user, min_rem, mail, config, ip)
        sys.exit()
    elif (
        not user.failed_login_attempts
        or user.failed_login_attempts >= max_failed_attempts
    ):
        # set failed login attempts to 1 if first failed login attempt or if
        # user got unblocked after being blocked previously
        user.failed_login_attempts = 1
    else:
        user.failed_login_attempts += 1
    user.last_login_attempt = dt.datetime.now(dt.timezone.utc)
    user.save()
    sys.exit()


def __notify_user_blocked(
    app: Flask, user: User, min_rem: int, mail: Mail, config: dict, ip: str
) -> None:
    """
    Sends an email to the user when their account is locked.

    This function also checks that emails are not sent too often to the same
    user.

    Parameters
    ----------
    app: flask.Flask
        The current Flask app
    user: :class:`~vantage6.server.model.user.User`
        User who is temporarily blocked
    min_rem: int
        Number of minutes remaining before the account is unlocked
    mail: flask_mail.Mail
        An instance of the Flask mail class. Used to send email to user in case
        of too many failed login attempts.
    config: dict
        Dictionary with configuration settings
    ip: str
        IP address from where the login attempt was made
    """
    log.info("User %s is locked. Sending them an email.", user.username)

    # check that email has not already been sent recently
    password_policy = config.get("password_policy", {})
    minutes_between_blocked_emails = password_policy.get(
        "between_user_emails_minutes",
        DEFAULT_BETWEEN_USER_EMAILS_MINUTES,
    )
    email_sent_recently = user.last_email_failed_login_sent and (
        dt.datetime.now(dt.timezone.utc)
        < user.last_email_failed_login_sent.replace(tzinfo=dt.timezone.utc)
        + dt.timedelta(minutes=minutes_between_blocked_emails)
    )
    if email_sent_recently:
        return

    # send email
    smtp_settings = config.get("smtp")
    if not smtp_settings:
        log.warning("No SMTP settings found in config - cannot send email!")
        return
    email_from = smtp_settings.get("email_from", DEFAULT_EMAIL_FROM_ADDRESS)
    support_email = config.get("support_email", DEFAULT_SUPPORT_EMAIL_ADDRESS)

    max_failed_attempts = password_policy.get(
        "max_failed_attempts", DEFAULT_MAX_FAILED_ATTEMPTS
    )
    template_vars = {
        "firstname": user.firstname if user.firstname else user.username,
        "number_of_allowed_attempts": max_failed_attempts,
        "ip": ip,
        "time": dt.datetime.now(dt.timezone.utc),
        "time_remaining": min_rem,
        "support_email": support_email,
    }

    with app.app_context():
        mail.send_email(
            "Failed login attempts on your vantage6 account",
            sender=email_from,
            recipients=[user.email],
            text_body=render_template("mail/blocked_account.txt", **template_vars),
            html_body=render_template("mail/blocked_account.html", **template_vars),
        )

    # Update latest email sent timestamp
    user.last_email_failed_login_sent = dt.datetime.now(dt.timezone.utc)
    user.save()


def create_qr_uri(user: User, server_name: str) -> dict:
    """
    Create the URI to generate a QR code for authenticator apps

    Parameters
    ----------
    user: :class:`~vantage6.server.model.user.User`
        User for whom two-factor authentication is to be set up
    server_name: str
        Name of the server. This is used in the issuer name in the URI

    Returns
    -------
    dict
        Dictionary with information on the TOTP secret required to generate
        a QR code or to enter it manually in an authenticator app
    """
    otp_secret = pyotp.random_base32()
    qr_uri = pyotp.totp.TOTP(otp_secret).provisioning_uri(
        name=user.username, issuer_name=f"{APPNAME} ({server_name})"
    )
    user.otp_secret = otp_secret
    user.save()
    return {
        "qr_uri": qr_uri,
        "otp_secret": otp_secret,
        "msg": (
            "Two-factor authentication is obligatory on this server. "
            "Please visualize the QR code to set up authentication."
        ),
    }



import logging

from http import HTTPStatus
from flask.globals import request
from flask import g
from flask_restful import Api
from sqlalchemy import or_

from vantage6.server.resource import with_user, ServicesResources
from vantage6.common import logger_name
from vantage6.server import db
from vantage6.server.resource.common.output_schema import RuleSchema
from vantage6.backend.common.resource.pagination import Pagination


module_name = logger_name(__name__)
log = logging.getLogger(module_name)
rule_schema = RuleSchema()


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the rule resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        Rules,
        path,
        endpoint="rule_without_id",
        methods=("GET",),
        resource_class_kwargs=services,
    )
    api.add_resource(
        Rule,
        path + "/<int:id>",
        endpoint="rule_with_id",
        methods=("GET",),
        resource_class_kwargs=services,
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class Rules(ServicesResources):
    @with_user
    def get(self):
        """List Rules
        ---
        description: >-
            List of all available rules at the server. The user must be
            authenticated, but does not require any additional permissions to
            view the rules.\n

            Accesible to users.

        parameters:
            - in: query
              name: name
              schema:
                type: string
              description: Filter by name of the rule
            - in: query
              name: operation
              schema:
                type: string
              description: Get rules for a specific type of operation
            - in: query
              name: scope
              schema:
                type: string
              description: Get rules for a specific scope
            - in: query
              name: role_id
              schema:
                type: integer
              description: Get rules for a specific role
            - in: query
              name: user_id
              schema:
                type: integer
              description: Get rules for a specific user. This includes the
                rules that are part of the user's roles.
            - in: query
              name: page
              schema:
                type: integer
              description: Page number for pagination (default=1)
            - in: query
              name: per_page
              schema:
                type: integer
              description: Number of items per page (default=10)
            - in: query
              name: sort
              schema:
                type: string
              description: >-
                Sort by one or more fields, separated by a comma. Use a minus
                sign (-) in front of the field to sort in descending order.
            - in: query
              name: no_pagination
              schema:
                type: int
              description: >-
                If set to 1, pagination is disabled and all items are returned.

        responses:
          200:
            description: Ok
          400:
            description: Improper values for pagination or sorting parameters

        security:
            - bearerAuth: []

        tags: ["Rule"]
        """
        q = g.session.query(db.Rule)

        args = request.args

        # filter by any field of this endpoint
        for param in ["name", "operation", "scope"]:
            if param in args:
                q = q.filter(getattr(db.Rule, param) == args[param])

        # find roles containing a specific rule
        if "role_id" in args:
            role = db.Role.get(args["role_id"])
            if not role:
                return {
                    "msg": f'Role with id={args["role_id"]} does not exist!'
                }, HTTPStatus.BAD_REQUEST
            q = (
                q.join(db.role_rule_association)
                .join(db.Role)
                .filter(db.Role.id == args["role_id"])
            )

        # find all rules of a specific user. This is done by first joining all
        # tables to find all rules originating from a user's roles. Then, we
        # do an outer join to find all rules that are directly assigned to the
        # user.
        if "user_id" in args:
            user = db.User.get(args["user_id"])
            if not user:
                return {
                    "msg": f'User with id={args["user_id"]} does not exist!'
                }, HTTPStatus.BAD_REQUEST
            q = (
                q.join(db.role_rule_association)
                .join(db.Role)
                .join(db.Permission)
                .join(db.User)
                .outerjoin(db.UserPermission, db.Rule.id == db.UserPermission.c.rule_id)
                .filter(
                    or_(
                        db.User.id == args["user_id"],
                        db.UserPermission.c.user_id == args["user_id"],
                    )
                )
            )

        # check if pagination is disabled
        paginate = True
        if "no_pagination" in args and args["no_pagination"] == "1":
            paginate = False

        # paginate results
        try:
            page = Pagination.from_query(q, request, db.Rule, paginate=paginate)
        except (ValueError, AttributeError) as e:
            return {"msg": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR

        # model serialization
        return self.response(page, rule_schema)


class Rule(ServicesResources):
    @with_user
    def get(self, id):
        """Returns a specific rule
        ---
        description: >-
            Get a rule by it's id. The user must be authenticated, but does
            not require any additional permissions to view rules.\n

            Accesible to users.

        parameters:
        - in: path
          name: id
          schema:
              type: integer
          minimum: 1
          description: rule_id
          required: true

        responses:
          200:
            description: Ok
          404:
            description: Rule not found

        security:
            - bearerAuth: []

        tags: ["Rule"]
        """
        rule = db.Rule.get(id)
        if not rule:
            return {"msg": f"Rule id={id} not found!"}, HTTPStatus.NOT_FOUND

        return rule_schema.dump(rule, many=False), HTTPStatus.OK



# TODO this file is awkward...
from vantage6.server.model import (
    Base,
    AlgorithmPort,
    AlgorithmStore,
    Authenticatable,
    Collaboration,
    Member,
    Node,
    NodeConfig,
    Organization,
    Permission,
    Role,
    Rule,
    Run,
    Study,
    StudyMember,
    Task,
    TaskDatabase,
    User,
    UserPermission,
    role_rule_association,
)



import logging

from http import HTTPStatus
from flask.globals import request
from flask import g
from flask_restful import Api
from sqlalchemy import or_

from vantage6.backend.common.resource.error_handling import (
    BadRequestError,
    NotFoundError,
    UnauthorizedError,
    handle_exceptions,
)
from vantage6.backend.common.resource.input_schema import ServerRoleInputSchema
from vantage6.backend.common.resource.role import (
    apply_user_filter,
    can_delete_dependents,
    check_default_role,
    filter_by_attribute,
    get_role,
    get_rule,
    get_rules,
    update_role,
    validate_request_body,
    validate_user_exists,
)
from vantage6.server import db
from vantage6.server.resource import (
    get_org_ids_from_collabs,
    with_user,
    ServicesResources,
)
from vantage6.common import logger_name
from vantage6.server.permission import (
    PermissionManager,
    RuleCollection,
    Operation as P,
)
from vantage6.server.model.rule import Operation, Scope
from vantage6.server.resource.common.output_schema import RoleSchema, RuleSchema
from vantage6.backend.common.resource.pagination import Pagination
from vantage6.server.default_roles import DefaultRole

module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the role resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        Roles,
        path,
        endpoint="role_without_id",
        methods=("GET", "POST"),
        resource_class_kwargs=services,
    )
    api.add_resource(
        Role,
        path + "/<int:id>",
        endpoint="role_with_id",
        methods=("GET", "PATCH", "DELETE"),
        resource_class_kwargs=services,
    )
    api.add_resource(
        RoleRules,
        path + "/<int:id>/rule/<int:rule_id>",
        endpoint="role_rule_with_id",
        methods=("DELETE", "POST"),
        resource_class_kwargs=services,
    )


# -----------------------------------------------------------------------------
# Permissions
# -----------------------------------------------------------------------------
def permissions(permissions: PermissionManager) -> None:
    """
    Define the permissions for this resource.

    Parameters
    ----------
    permissions : PermissionManager
        Permission manager instance to which permissions are added
    """
    add = permissions.appender(module_name)
    add(scope=Scope.GLOBAL, operation=Operation.VIEW, description="View any role")
    add(
        scope=Scope.COLLABORATION,
        operation=Operation.VIEW,
        description="View any role in your collaborations",
    )
    add(
        scope=Scope.ORGANIZATION,
        operation=Operation.VIEW,
        description="View the roles of your organization",
    )
    add(
        scope=Scope.GLOBAL,
        operation=Operation.CREATE,
        description="Create role for any organization",
    )
    add(
        scope=Scope.COLLABORATION,
        operation=Operation.CREATE,
        description="Create role for any organization in your collaborations",
    )
    add(
        scope=Scope.ORGANIZATION,
        operation=Operation.CREATE,
        description="Create role for your organization",
    )
    add(scope=Scope.GLOBAL, operation=Operation.EDIT, description="Edit any role")
    add(
        scope=Scope.COLLABORATION,
        operation=Operation.EDIT,
        description="Edit any role in your collaborations",
    )
    add(
        scope=Scope.ORGANIZATION,
        operation=Operation.EDIT,
        description="Edit a role from your organization",
    )
    add(
        scope=Scope.GLOBAL,
        operation=Operation.DELETE,
        description="Delete a role from any organization",
    )
    add(
        scope=Scope.COLLABORATION,
        operation=Operation.DELETE,
        description="Delete a role from any organization in your " "collaborations",
    )
    add(
        scope=Scope.ORGANIZATION,
        operation=Operation.DELETE,
        description="Delete a role from your organization",
    )


# -----------------------------------------------------------------------------
# Resources / API's
# -----------------------------------------------------------------------------
role_schema = RoleSchema()
rule_schema = RuleSchema()
role_input_schema = ServerRoleInputSchema(default_roles=DefaultRole.list())


class RoleBase(ServicesResources):
    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        self.rule_collection: RuleCollection = getattr(self.permissions, module_name)

    def _get_organization_id(self, data):
        return data.get("organization_id", g.user.organization_id)

    def _validate_organization(self, organization_id):
        if not db.Organization.get(organization_id):
            raise NotFoundError(f'Organization "{organization_id}" does not exist!')

    def _validate_user_permission(self, operation, organization_id):
        if not self.rule_collection.allowed_for_org(operation, organization_id):
            raise UnauthorizedError(
                f"You lack the permission to {operation} roles for organization {organization_id}!"
            )


class Roles(RoleBase):
    def _filter_by_organization(self, query, args):
        org_filters = args.getlist("organization_id")
        if org_filters:
            for org_id in org_filters:
                if not self.rule_collection.allowed_for_org(P.VIEW, org_id):
                    raise UnauthorizedError(
                        f"You lack the permission to view roles from organization {org_id}!"
                    )
            include_root = args.get("include_root", False)
            query = query.filter(
                or_(
                    db.Role.organization_id.in_(org_filters),
                    db.Role.organization_id.is_(None) if include_root else False,
                )
            )
        return query

    def _filter_by_collaboration(self, query, args):
        if "collaboration_id" in args:
            if not self.rule_collection.can_for_col(P.VIEW, args["collaboration_id"]):
                raise UnauthorizedError(
                    f"You lack the permission to view roles from collaboration {args['collaboration_id']}!"
                )
            org_ids = get_org_ids_from_collabs(g.user, args["collaboration_id"])
            include_root = args.get("include_root", False)
            query = query.filter(
                or_(
                    db.Role.organization_id.in_(org_ids),
                    db.Role.organization_id.is_(None) if include_root else False,
                )
            )
        return query

    def _filter_by_rule(self, query, args):
        if "rule_id" in args:
            get_rule(db, args["rule_id"])
            query = (
                query.join(db.role_rule_association)
                .join(db.Rule)
                .filter(db.Rule.id == args["rule_id"])
            )
        return query

    def _filter_by_user(self, query, args):
        if "user_id" in args:
            user = validate_user_exists(db, args["user_id"])
            if (
                not self.rule_collection.allowed_for_org(P.VIEW, user.organization_id)
                and not g.user.id == user.id
            ):
                raise UnauthorizedError(
                    f"You lack the permission to view roles from the organization that user id={user.id} belongs to!"
                )
            query = apply_user_filter(db, query, args["user_id"])
        return query

    def _filter_by_user_permissions(self, query, auth_org):
        own_role_ids = [role.id for role in g.user.roles]
        if self.rule_collection.v_col.can():
            query = query.filter(
                or_(
                    db.Role.id.in_(own_role_ids),
                    db.Role.organization_id.is_(None),
                    db.Role.organization_id.in_(
                        [
                            org.id
                            for col in self.obtain_auth_collaborations()
                            for org in col.organizations
                        ]
                    ),
                )
            )
        elif self.rule_collection.v_org.can():
            query = query.filter(
                or_(
                    db.Role.organization_id == auth_org.id,
                    db.Role.id.in_(own_role_ids),
                    db.Role.organization_id.is_(None),
                )
            )
        else:
            query = query.filter(db.Role.id.in_(own_role_ids))
        return query

    @with_user
    @handle_exceptions
    def get(self):
        """Returns a list of roles
        ---

        description: >-
            Returns a list of roles. Depending on your permission, you get all
            the roles at the server or only the roles that belong to your
            organization.\n

            ### Permission Table\n
            |Rule name|Scope|Operation|Assigned to node|Assigned to container|
            Description|\n
            |--|--|--|--|--|--|\n
            |Role|Global|View|❌|❌|View all roles|\n
            |Role|Collaboration|View|❌|❌|View all roles in your
            collaborations|\n
            |Role|Organization|View|❌|❌|View roles that are part of your
            organization|\n

            Accesible to users.

        parameters:
            - in: query
              name: name
              schema:
                type: string
              description: >-
                Name to match with a LIKE operator. \n
                * The percent sign (%) represents zero, one, or multiple
                characters\n
                * underscore sign (_) represents one, single character
            - in: query
              name: description
              schema:
                type: string
              description: >-
                Description to match with a LIKE operator. \n
                * The percent sign (%) represents zero, one, or multiple
                characters\n
                * underscore sign (_) represents one, single character
            - in: query
              name: organization_id
              schema:
                type: array
                items:
                  type: integer
                  description: Organization id of which you want to get roles
            - in: query
              name: collaboration_id
              schema:
              type: integer
              description: Collaboration id
            - in: query
              name: rule_id
              schema:
                type: integer
              description: Get roles that this role id is part of
            - in: query
              name: user_id
              schema:
                type: integer
              description: Get roles for this user id
            - in: query
              name: include_root
              schema:
                 type: boolean
              description: Whether or not to include root role (default=False)
            - in: query
              name: page
              schema:
                type: integer
              description: Page number for pagination (default=1)
            - in: query
              name: per_page
              schema:
                type: integer
              description: Number of items per page (default=10)
            - in: query
              name: sort
              schema:
                type: string
              description: >-
                Sort by one or more fields, separated by a comma. Use a minus
                sign (-) in front of the field to sort in descending order.

        responses:
          200:
            description: Ok
          401:
            description: Unauthorized
          400:
            description: Improper values for pagination or sorting parameters

        security:
            - bearerAuth: []

        tags: ["Role"]
        """
        q = g.session.query(db.Role)
        auth_org = self.obtain_auth_organization()

        q = self._filter_by_organization(q, request.args)
        q = self._filter_by_collaboration(q, request.args)
        q = filter_by_attribute(db, ["name", "description"], q, request.args)
        q = self._filter_by_rule(q, request.args)
        q = self._filter_by_user(q, request.args)

        if not self.rule_collection.v_glo.can():
            q = self._filter_by_user_permissions(q, auth_org)

        page = Pagination.from_query(q, request, db.Role)
        return self.response(page, role_schema)

    @with_user
    @handle_exceptions
    def post(self):
        """Creates a new role.
        ---
        description: >-
          Create a new role. You can only assign rules that you own. You need
          permission to create roles, and you can only assign roles to other
          organizations if you have gobal permission.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Role|Global|Create|❌|❌|Create a role for any organization|\n
          |Role|Collaboration|Create|❌|❌|Create a role for organization in
          your collaborations|\n
          |Role|Organization|Create|❌|❌|Create a role for your organization|\n

          Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  name:
                    type: string
                    description: Human readable name for collaboration
                  description:
                    type: string
                    description: Human readable description of the role
                  rules:
                    type: array
                    items:
                      type: integer
                      description: Rule id's to assign to role
                  organization_id:
                    type: integer
                    description: Organization to which role is added. If you
                      are root user and want to create a role that will be
                      available to all organizations, leave this empty.

        responses:
          201:
            description: Created
          400:
            description: Non-allowed role name
          401:
            description: Unauthorized
          404:
            description: Organization or rule was not found

        security:
          - bearerAuth: []

        tags: ["Role"]
        """
        data = request.get_json()
        validate_request_body(role_input_schema, data)

        rules = get_rules(data, db)
        self.permissions.check_user_rules(rules)
        organization_id = self._get_organization_id(data)
        self._validate_organization(organization_id)
        self._validate_user_permission(P.CREATE, organization_id)

        role = db.Role(
            name=data.get("name"),
            description=data.get("description"),
            rules=rules,
            organization_id=organization_id,
        )
        role.save()

        return role_schema.dump(role, many=False), HTTPStatus.CREATED


class Role(RoleBase):
    def has_permission_to_view(self, role) -> bool:
        return (
            self.rule_collection.allowed_for_org(P.VIEW, role.organization_id)
            or role in g.user.roles
            or (
                role.name in [role for role in DefaultRole]
                and self.rule_collection.has_at_least_scope(Scope.ORGANIZATION, P.VIEW)
            )
        )

    @with_user
    @handle_exceptions
    def get(self, id):
        """Get roles
        ---
        description: >-
          Get role based on role identifier.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Role|Global|View|❌|❌|View all roles|\n
          |Role|Collaboration|View|❌|❌|View all roles for your
          collaborations|\n
          |Role|Organization|View|❌|❌|View roles that are part of your
          organization|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
              minimum: 1
            description: Role id
            required: true

        responses:
          200:
            description: Ok
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Role"]
        """
        role = get_role(db, id)

        if not self.has_permission_to_view(role):
            raise UnauthorizedError("You do not have permission to view this role.")

        return role_schema.dump(role, many=False), HTTPStatus.OK

    @with_user
    @handle_exceptions
    def patch(self, id):
        """Update role
        ---
        description: >-
          Updates roles if the user has permission to do so.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Role|Global|Edit|❌|❌|Update any role|\n
          |Role|Collaboration|Edit|❌|❌|Update any role in your
          collaborations|\n
          |Role|Organization|Edit|❌|❌|Update a role from your organization|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
              minimum: 1
            description: Role id
            required: tr

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  name:
                    type: string
                    description: Human readable name for collaboration
                  description:
                    type: string
                    description: Human readable description of the role
                  rules:
                    type: array
                    items:
                      type: integer
                      description: Rule id's to assign to role

        responses:
          200:
            description: Ok
          400:
            description: Non-allowed role name change
          401:
            description: Unauthorized
          404:
            description: Role id or rule id not found

        security:
          - bearerAuth: []

        tags: ["Role"]
        """
        data = request.get_json()
        validate_request_body(role_input_schema, data, partial=True)

        if "organization_id" in data:
            raise BadRequestError("Cannot change role's organization")

        role = get_role(db, id)
        check_default_role(role, DefaultRole.list())
        self._validate_user_permission(P.EDIT, role.organization_id)
        role = update_role(role, data, db, self.permissions)
        role.save()

        return role_schema.dump(role, many=False), HTTPStatus.OK

    @with_user
    @handle_exceptions
    def delete(self, id):
        """Delete role
        ---
        description: >-
          Delete role from an organization if user is allowed to delete the
          role.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Role|Global|Delete|❌|❌|Delete any role|\n
          |Role|Collaboration|Delete|❌|❌|Delete any role in your
          collaborations|\n
          |Role|Organization|Delete|❌|❌|Delete a role in your organization|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
              minimum: 1
            description: Role id
            required: true
          - in: query
            name: delete_dependents
            schema:
              type: boolean
            description: If set to true, the role is deleted even though users
               with this role may lose permissions (default=False)

        responses:
          200:
            description: Ok
          400:
            description: Cannot delete default roles
          401:
            description: Unauthorized
          404:
            description: Role id not found

        security:
          - bearerAuth: []

        tags: ["Role"]
        """
        role = get_role(db, id)
        check_default_role(role, DefaultRole.list())
        self._validate_user_permission(P.DELETE, role.organization_id)
        can_delete_dependents(role, request.args.get("delete_dependents", False))
        role.delete()

        return {"msg": "Role removed from the database."}, HTTPStatus.OK


class RoleRules(RoleBase):
    @with_user
    @handle_exceptions
    def post(self, id, rule_id):
        """Add a rule to a role.
        ---
        description: >-
          Add a rule to a role given that the role exists already and that the
          user has the permission to do so.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Role|Global|Edit|❌|❌|Edit any role|\n
          |Role|Collaboration|Edit|❌|❌|Edit any role in your collaborations
          |\n
          |Role|Organization|Edit|❌|❌|Edit any role in your organization|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Role id
            required: tr
          - in: path
            name: rule_id
            schema:
              type: integer
            description: Rule id to add to role
            required: tr

        responses:
          201:
            description: Added rule to role
          404:
            description: Rule or role not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Role"]
        """
        role = get_role(db, id)
        rule = get_rule(db, rule_id)
        self._validate_user_permission(P.EDIT, role.organization_id)
        self.permissions.check_user_rules([rule])
        role.rules.append(rule)
        role.save()

        return rule_schema.dump(role.rules, many=True), HTTPStatus.CREATED

    @with_user
    @handle_exceptions
    def delete(self, id, rule_id):
        """Removes rule from role.
        ---
        description: >-
          Removes a rule from a role given the user has permission and the rule
          id exists.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Role|Global|Edit|❌|❌|Delete any rule in a role|\n
          |Role|Collaboration|Edit|❌|❌|Delete any rule in roles in your
          collaborations|\n
          |Role|Organization|Edit|❌|❌|Delete any rule in roles in your
          organization|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Role id
          - in: path
            name: rule_id
            schema:
              type: integer
            description: Rule id to delete from the role
            required: true

        responses:
          200:
            description: Ok
          404:
            description: Role or rule id not found
          401:
            description: Unauthorized

        tags: ["Role"]
        """
        role = get_role(db, id)
        rule = get_rule(db, rule_id)
        self._validate_user_permission(P.EDIT, role.organization_id)
        self.permissions.check_user_rules([rule])
        if rule not in role.rules:
            raise NotFoundError(
                f"Rule with id={rule_id} is not part of role with id={id}"
            )
        role.rules.remove(rule)
        role.save()

        return rule_schema.dump(role.rules, many=True), HTTPStatus.OK



"""
Database connection classes for the server.

Note that this file is almost identical to the one for the algorithm store. However,
this is necessary because the database class is a singleton so we have to create
two different classes for it.
"""

import logging
from typing import Any

from sqlalchemy.orm.session import Session
from sqlalchemy.ext.declarative import declarative_base

from vantage6.common import logger_name, Singleton
from vantage6.backend.common.base import (
    BaseModelBase,
    BaseDatabaseSessionManager,
    BaseDatabase,
)

module_name = logger_name(__name__)
log = logging.getLogger(module_name)


class Database(BaseDatabase, metaclass=Singleton):
    """
    Database class that is used to connect to the database and create the
    database session.

    The database is created as a singleton, so that it can be destroyed (as
    opposed to a module). This is especially useful when creating unit tests
    in which we want fresh databases every now and then.
    """

    def drop_all(self):
        """
        Drop all tables in the database.
        """
        self._drop_all(Base)

    def clear_data(self):
        """
        Clear all data from the database.
        """
        self._clear_data(Base, DatabaseSessionManager)

    def close(self):
        """
        Delete all tables and close the database connection. Only used for
        unit testing.
        """
        self._close(Base)

    def connect(self, uri="sqlite:////tmp/test.db", allow_drop_all=False):
        """
        Connect to the database.

        Parameters
        ----------
        uri : str
            URI of the database. Defaults to a sqlite database in /tmp.
        allow_drop_all : bool, optional
            If True, the database can be dropped. Defaults to False.
        """
        self._connect(Base, uri, allow_drop_all)


class DatabaseSessionManager(BaseDatabaseSessionManager):
    """
    Class to manage DB sessions.

    There are 2 different ways a session can be obtained. Either a session used
    within a request or a session used elsewhere (e.g. socketIO event, iPython
    or within the application itself).

    In case of the Flask request, the session is stored in the flask global
    `g`. Then, it can be accessed in every endpoint.

    In all other cases the session is attached to the db module.
    """

    @staticmethod
    def get_session() -> Session:
        """
        Get a session. Creates a new session if none exists.

        Returns
        -------
        Session
            A database session
        """
        return BaseDatabaseSessionManager._get_session(DatabaseSessionManager)

    @staticmethod
    def new_session() -> None:
        """
        Create a new session. If we are in a flask request, the session is
        stored in the flask global `g`. Otherwise, the session is stored in
        the db module.
        """
        BaseDatabaseSessionManager._new_session(DatabaseSessionManager, Database)


class ModelBase(BaseModelBase):
    """
    Declarative base that defines default attributes. All data models inherit
    from this class.
    """

    @classmethod
    def get(cls, id_: int = None):
        """
        Get a single object by its id, or a list of objects when no id is
        specified.

        Parameters
        ----------
        id_: int, optional
            The id of the object to get. If not specified, return all.
        """
        return cls._get(DatabaseSessionManager, id_)

    def save(self) -> None:
        """
        Save the object to the database.
        """
        return self._save(DatabaseSessionManager)

    def delete(self) -> None:
        """
        Delete the object from the database.
        """
        return self._delete(DatabaseSessionManager)

    @classmethod
    def exists(cls, field: str, value: Any) -> bool:
        """
        Check if a value exists for a given field in the database model.

        Parameters
        ----------
        field: str
            The field to check
        value: Any
            The value to check

        Returns
        -------
        bool
            True if the value exists, False otherwise
        """
        return cls._exists(field, value, DatabaseSessionManager)


Base = declarative_base(cls=ModelBase)



import logging
import uuid

# uwsgi is not available when running outside of a uwsgi process.
# (see https://uwsgi-docs.readthedocs.io/en/latest/PythonModule.html)
# This prevents a ModuleNotFoundError if this resource is loaded
# outside of such a server, e.g. when importing data in v6 dev create-demo-network.
try:
    import uwsgi  # type: ignore (built from source in Dockerfile)
except ImportError:
    uwsgi = None
from flask import g, request, Response, stream_with_context

from flask_restful import Api
from http import HTTPStatus

from vantage6.common import logger_name
from vantage6.common.task_status import has_task_finished
from vantage6.server.permission import RuleCollection, Operation as P, Scope
from vantage6.server.resource import (
    only_for,
    ServicesResources,
)
from vantage6.server.model import Run as db_Run, Task as db_Task


module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the run resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        BlobStream,
        api_base + "/blobstream",
        endpoint="result_stream_without_id",
        methods=("POST",),
        resource_class_kwargs=services,
    )

    api.add_resource(
        BlobStreamStatus,
        api_base + "/blobstream/status",
        endpoint="blob_stream_status",
        methods=("GET",),
        resource_class_kwargs=services,
    )

    api.add_resource(
        BlobStream,
        api_base + "/blobstream/<string:id>",
        endpoint="blob_stream_with_id",
        methods=("GET",),
        resource_class_kwargs=services,
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class BlobStreamBase(ServicesResources):
    """
    Base class for run resources that require a storage adapter.
    This class provides methods for streaming large results from the storage adapter.
    """

    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        self.r_run: RuleCollection = getattr(self.permissions, "run")
        self.r_task: RuleCollection = getattr(self.permissions, "task")
        self.storage_adapter = storage_adapter

    def get_run_by_input_or_result(self, id) -> db_Run | None:
        run = (
            g.session.query(db_Run)
            .filter((db_Run.input == id) | (db_Run.result == id))
            .first()
        )
        return run


class BlobStreamStatus(BlobStreamBase):
    """
    Resource for /api/blobstream/status (GET)
    Returns whether the blob store is enabled.
    """

    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)

    @only_for(("node", "user", "container"))
    def get(self):
        """Get the status of the blob store
        ---

        description: >-
            Returns whether or not blob storage is enabled. \n

        responses:
          200:
              description: Ok
          401:
              description: Unauthorized

        security:
          - bearerAuth: []
        """
        log.debug("Checking if blob store is enabled")

        if self.storage_adapter:
            return {"blob_store_enabled": True}, HTTPStatus.OK
        else:
            return {"blob_store_enabled": False}, HTTPStatus.OK


class BlobStream(BlobStreamBase):
    """
    Resource for /api/blobstream/<id> (GET) and /api/blobstream (POST)
    This resource allows for streaming large results from blob Storage.
    """

    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)

    @only_for(("node", "user", "container"))
    def get(self, id):
        """Stream the result or input with the given id from blob storage.
        ---

        description: >-
            Streams the result or input with the given id from blob storage.

            ### Permission Table\n
            |Rule name|Scope|Operation|Assigned to node|Assigned to container|
            Description|\n
            |--|--|--|--|--|--|\n
            |Blobstream|Global|View|❌|❌|View any blob|\n
            |Blobstream|Collaboration|View|✅|✅|View the blobs of your
            organization's collaborations|\n
            |Blobstream|Organization|View|❌|❌|View any blob from a task created by
            your organization|\n
            |Blobstream|Own|View|❌|❌|View any blob from a task created by you|\n

            Accessible to users.

        responses:
          200:
            description: Ok
          401:
            description: Unauthorized
          404:
            description: Not Found
          500:
            description: Internal Server Error
          501:
            description: Not Implemented

        security:
        - bearerAuth: []

        tags: ["Algorithm"]
        """
        run = self.get_run_by_input_or_result(id)
        if not run:
            return {
                "msg": f"No run found with input or result id={id}"
            }, HTTPStatus.NOT_FOUND
        if not self.r_run.allowed_for_org(P.VIEW, run.task.init_org_id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        if not self.storage_adapter:
            not_available_msg = "The large result store is not set to blob storage, result streaming is not available."
            log.warning(not_available_msg)
            return {"msg": not_available_msg}, HTTPStatus.NOT_IMPLEMENTED
        try:
            log.debug(f"Streaming result for run id={id}")
            blob_stream = self.storage_adapter.stream_blob(id)
        except Exception as e:
            log.error(f"Error streaming result: {e}")
            return {"msg": "Error streaming result!"}, HTTPStatus.INTERNAL_SERVER_ERROR

        def generate():
            for chunk in blob_stream.chunks():
                yield chunk

        return Response(
            stream_with_context(generate()),
            content_type="application/octet-stream",
            headers={"Content-Disposition": f"attachment; filename=result_{id}.bin"},
        )

    @only_for(("node", "user", "container"))
    def post(self):
        """Post a result to the blob storage.
        blobs are streamed directly to the blob storage.
        This is useful for large results that cannot be loaded into memory at once.
        ---
        description: >-
          Stream and store blob to blob storage.

        requestBody:
          content:
            application/octet-stream

        responses:
          201:
            description: Created
          501:
            description: Not Implemented

        security:
          - bearerAuth: []

        tags: ["Algorithm"]
        """
        if not self.storage_adapter:
            not_available_msg = "The large result store is not set to blob storage, result streaming is not available."
            log.warning(not_available_msg)
            return {"msg": not_available_msg}, HTTPStatus.NOT_IMPLEMENTED

        if g.user and not self.r_task.has_at_least_scope(Scope.COLLABORATION, P.CREATE):
            return {
                "msg": "You do not have permission to upload blobs. This requires permission to create tasks which you don't have."
            }, HTTPStatus.UNAUTHORIZED
        if g.container:
            container = g.container
            if has_task_finished(db_Task.get(container["task_id"]).status):
                log.warning(
                    f"Container from node={container['node_id']} "
                    f"attempts to upload blob for a sub-task of a completed "
                    f"task={container['task_id']}"
                )
                return {
                    "msg": "Cannot upload blob for a sub-task of a completed task."
                }, HTTPStatus.FORBIDDEN

        result_uuid = str(uuid.uuid4())
        transfer_encoding = request.headers.get("Transfer-Encoding", "").lower()
        is_chunked = "chunked" in transfer_encoding
        try:
            # Unfortunately, in the case of streams smaller than one chunk,
            # reverse proxies like nginx will automatically remove the chunked transfer encoding.
            # Therefore, we need to handle both chunked and non-chunked uploads. Exchanging uwsgi
            # to a different server might solve this issue.
            if is_chunked:
                stream = UwsgiChunkedStream()
                self.storage_adapter.store_blob(result_uuid, stream)
            else:
                data = request.get_data()
                self.storage_adapter.store_blob(result_uuid, data)
        except Exception as e:
            log.error(f"Error uploading result: {e}")
            return {"msg": "Error uploading result!"}, HTTPStatus.INTERNAL_SERVER_ERROR

        return {"uuid": result_uuid}, HTTPStatus.CREATED


class UwsgiChunkedStream:
    """
    Read data in chunks from uwsgi.
    """

    # TODO: Using uwsgi in python in combination with flask is not ideal.
    # It would be better to switch to a different server in the long term.
    #
    def __init__(self, chunk_size=4096):
        """
        Initialize the UwsgiChunkedStream.
        """
        self.chunk_size = chunk_size
        self._buffer = b""
        self._eof = False

    def read(self, size=-1):
        """
        Read data from the stream.
        """
        log.info(f"Reading with size: {size}")
        if size < 0:
            chunks = [self._buffer]
            self._buffer = b""
            while not self._eof:
                chunk = uwsgi.chunked_read(self.chunk_size)
                if not chunk:
                    self._eof = True
                    break
                chunks.append(chunk)
            return b"".join(chunks)

        while len(self._buffer) < size and not self._eof:
            chunk = uwsgi.chunked_read(self.chunk_size)
            if not chunk:
                self._eof = True
                break
            self._buffer += chunk

        result, self._buffer = self._buffer[:size], self._buffer[size:]
        return result



import uuid
import logging

import vantage6.server.model as db
from vantage6.server.model.base import Database
from vantage6.server import RESOURCES, RESOURCES_PATH, DefaultRole
from vantage6.server.permission import PermissionManager
from vantage6.common.task_status import TaskStatus
from vantage6.common.serialization import serialize
from vantage6.common import bytes_to_base64s

module_name = __name__.split(".")[-1]
log = logging.getLogger(module_name)


# TODO maybe move this function to a more general location (utils?)
def _is_valid_uuid(uuid_to_test):
    """
    Check if uuid_to_test is a valid UUID.

     Parameters
    ----------
    uuid_to_test : str

     Returns
    -------
    `True` if uuid_to_test is a valid UUID, otherwise `False`.
    """
    try:
        uuid_obj = uuid.UUID(uuid_to_test)
    except ValueError:
        return False
    return str(uuid_obj) == uuid_to_test


def load(fixtures: dict, drop_all: bool = False) -> None:
    """
    Load fixtures (i.e. fixed resources to test functionality) into the
    database.

    Parameters
    ----------
    fixtures : dict
        Dictionary containing the fixtures to load into the database.
    drop_all : bool
        If `True` all tables in the database will be dropped before loading
    """
    # TODO we are not sure the DB is connected here....

    if drop_all:
        Database().drop_all()

    # If the server has never been started yet, no rules have been created yet.
    # In that case, create them here so that the created users have
    # permissions.
    if not db.Rule.get():
        permissions = PermissionManager(RESOURCES_PATH, RESOURCES, DefaultRole)
        permissions.load_rules_from_resources(RESOURCES_PATH, RESOURCES)

    log.info("Create Organizations and Users")
    for org in fixtures.get("organizations", {}):
        # print(org)

        # create organization
        organization = db.Organization(
            **{
                k: org[k]
                for k in [
                    "name",
                    "domain",
                    "address1",
                    "address2",
                    "zipcode",
                    "country",
                    "public_key",
                ]
            }
        )
        organization.save()
        log.debug(f"processed organization={organization.name}")
        superuserrole = db.Role(
            name="super",
            description="Super user",
            rules=db.Rule.get(),
            organization=organization,
        )
        superuserrole.save()
        # create users
        for usr in org.get("users", {}):
            user = db.User(**usr)
            user.roles = [superuserrole]
            user.organization = organization
            user.save()
            log.debug(f"processed user={user.username}")

    log.info("Create collaborations")
    for col in fixtures.get("collaborations", {}):
        # create collaboration
        collaboration = db.Collaboration(
            name=col.get("name"), encrypted=col.get("encrypted", True)
        )
        log.debug(f"processed collaboration={collaboration.name}")

        # append organizations to the collaboration

        for participant in col.get("participants", {}):
            if (
                not isinstance(participant, dict)
                or not participant.get("name")
                or not participant.get("api_key")
            ):
                log.error(
                    "Collaboration participants should contain the "
                    "fields 'name' and 'api_key'. This is not the case "
                    f"for participant {participant} in collaboration "
                    f"{collaboration.name}"
                )
                exit(1)

            org_name = participant.get("name")
            node_api_key = participant.get("api_key")

            # check if api key is valid uuid
            if not _is_valid_uuid(node_api_key):
                log.error(f"API key '{node_api_key}' is not a valid UUID!")
                exit(1)

            organization = db.Organization.get_by_name(org_name)
            collaboration.organizations.append(organization)
            log.debug(f"added {org_name} to the collaboration")

            node = db.Node(
                organization=organization,
                collaboration=collaboration,
                name=f"{organization.name} - {collaboration.name} Node",
                api_key=node_api_key,
            )
            node.save()
            log.debug(f"added node {node.name} to {collaboration.name}")

        collaboration.save()

        # append dummy tasks to the collaboration
        log.debug("Processing Task Assignments")
        for image in col.get("tasks", {}):
            init_org = collaboration.organizations[0]
            task = db.Task(
                name="Example task",
                image=image,
                collaboration=collaboration,
                job_id=db.Task.next_job_id(),
                init_org=init_org,
                init_user=db.User.get()[0],
            )

            for organization in collaboration.organizations:
                run = db.Run(
                    task=task,
                    input=bytes_to_base64s(serialize({"a": "b"})),
                    organization=organization,
                    status=TaskStatus.PENDING,
                )
                run.save()

            task.save()
            log.debug(f"Processed task {task.name}")
